"""
implementation of load cli command
"""

import sys
import gzip
import json
import requests
from datetime import datetime
import re
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from ckanapi.errors import (NotFound, NotAuthorized, ValidationError,
    SearchIndexError)
from ckanapi.cli import workers
from ckanapi.cli.utils import completion_stats, compact_json, quiet_int_pipe

try:
    unicode
except NameError:
    unicode = str


def load_things(ckan, thing, arguments,
        worker_pool=None, stdin=None, stdout=None, stderr=None):
    """
    create and update datasets, groups, orgs and users

    The parent process creates a pool of worker processes and hands
    out json lines to each worker as they finish a task. Status of
    last record completed and records being processed is displayed
    on stderr.
    """
    if worker_pool is None:
        worker_pool = workers.worker_pool
    if stdin is None:
        stdin = getattr(sys.stdin, 'buffer', sys.stdin)
    if stdout is None:
        stdout = getattr(sys.__stdout__, 'buffer', sys.__stdout__)
    if stderr is None:
        stderr = getattr(sys.stderr, 'buffer', sys.stderr)

    if arguments['--worker']:
        return load_things_worker(ckan, thing, arguments)

    log = None
    if arguments['--log']:
        log = open(arguments['--log'], 'a')

    jsonl_input = stdin
    if arguments['--input']:
        jsonl_input = open(arguments['--input'], 'rb')
    if arguments['--gzip']:
        jsonl_input = gzip.GzipFile(fileobj=jsonl_input)

    def line_reader():
        """
        handle start-record and max-records options
        """
        start_record = int(arguments['--start-record'])
        max_records = arguments['--max-records']
        if max_records is not None:
            max_records = int(max_records)
        for num, line in enumerate(jsonl_input, 1): # records start from 1
            if num < start_record:
                continue
            if max_records is not None and num >= start_record + max_records:
                break
            yield num, line

    cmd = _worker_command_line(thing, arguments)
    processes = int(arguments['--processes'])
    if hasattr(ckan, 'parallel_limit'):
        # add your sites to ckanapi.remoteckan.MY_SITES instead of removing
        processes = min(processes, ckan.parallel_limit)
    stats = completion_stats(processes)
    pool = worker_pool(cmd, processes, line_reader())

    with quiet_int_pipe() as errors:
        for job_ids, finished, result in pool:
            if not result:
                # child exited with traceback
                return 1
            timestamp, action, error, response = json.loads(
                result.decode('utf-8'))

            if not arguments['--quiet']:
                stderr.write(('%s %s %s %s %s %s\n' % (
                    finished,
                    job_ids,
                    next(stats),
                    action,
                    error,
                    compact_json(response).decode('utf-8') if response else ''
                    )).encode('utf-8'))

            if log:
                log.write(compact_json([
                    timestamp,
                    finished,
                    action,
                    error,
                    response,
                    ]) + b'\n')
                log.flush()
    if 'pipe' in errors:
        return 1
    if 'interrupt' in errors:
        return 2


def load_things_worker(ckan, thing, arguments,
        stdin=None, stdout=None):
    """
    a process that accepts lines of json on stdin which is parsed and
    passed to the {thing}_create/update actions.  it produces lines of json
    which are the responses from each action call.
    """
    if stdin is None:
        stdin = getattr(sys.stdin, 'buffer', sys.stdin)
        # hack so that pdb can be used in extension/ckan
        # code called by this worker
        try:
            sys.stdin = open('/dev/tty', 'rb')
        except IOError:
            pass
    if stdout is None:
        stdout = getattr(sys.__stdout__, 'buffer', sys.__stdout__)
        # hack so that "print debugging" can work in extension/ckan
        # code called by this worker
        sys.stdout = sys.stderr

    thing_show, thing_create, thing_update = {
        'datasets': (
            'package_show', 'package_create', 'package_update'),
        'groups': (
            'group_show', 'group_create', 'group_update'),
        'organizations': (
            'organization_show', 'organization_create', 'organization_update'),
        'users': (
            'user_show', 'user_create', 'user_update'),
        'related':(
            'related_show','related_create','related_update'),
        }[thing]

    def reply(action, error, response):
        """
        format messages to be sent back to parent process
        """
        stdout.write(compact_json([
            datetime.now().isoformat(),
            action,
            error,
            response]) + b'\n')
        stdout.flush()

    for line in iter(stdin.readline, b''):
        try:
            obj = json.loads(line.decode('utf-8'))
        except UnicodeDecodeError as e:
            obj = None
            reply('read', 'UnicodeDecodeError', unicode(e))
            continue

        if obj is not None:
            existing = None
            if not arguments['--create-only']:
                # use either id or name to locate existing records
                name = obj.get('id')
                if name:
                    try:
                        existing = ckan.call_action(thing_show,
                            {'id': name,
                             'include_datasets': False,
                             'include_password_hash': True,
                            })
                    except NotFound:
                        pass
                    except NotAuthorized as e:
                        reply('show', 'NotAuthorized', unicode(e))
                        continue
                name = obj.get('name')
                if not existing and name:
                    try:
                        existing = ckan.call_action(thing_show, {'id': name})
                    except NotFound:
                        pass
                    except NotAuthorized as e:
                        reply('show', 'NotAuthorized', unicode(e))
                        continue

                if existing:
                    _copy_from_existing_for_update(obj, existing, thing)

                # FIXME: compare and reply when 'unchanged'?

            if not existing and arguments['--update-only']:
                reply('show', 'NotFound', [obj.get('id'), obj.get('name')])
                continue

            act = 'update' if existing else 'create'
            try:
                if existing:
                    r = ckan.call_action(thing_update, obj)
                else:
                    r = ckan.call_action(thing_create, obj)
                if thing == 'datasets' and 'resources' in obj:# check if it is needed to upload resources when creating/updating packages
                    _upload_resources(ckan,obj,arguments)
                elif thing in ['groups','organizations'] and 'image_display_url' in obj:   #load images for groups and organizations
                    if arguments['--upload-logo']:
                        users = obj['users']
                        _upload_logo(ckan,obj)
                        obj.pop('image_upload')
                        obj['users'] = users
                        ckan.call_action(thing_update,obj)
            except ValidationError as e:
                reply(act, 'ValidationError', e.error_dict)
            except SearchIndexError as e:
                reply(act, 'SearchIndexError', unicode(e))
            except NotAuthorized as e:
                reply(act, 'NotAuthorized', unicode(e))
            except NotFound:
                reply(act, 'NotFound', obj)
            else:
                reply(act, None, r.get('name',r.get('id')))

def _worker_command_line(thing, arguments):
    """
    Create a worker command line suitable for Popen with only the
    options the worker process requires
    """
    def a(name):
        "options with values"
        return [name, arguments[name]] * (arguments[name] is not None)
    def b(name):
        "boolean options"
        return [name] * bool(arguments[name])
    return (
        ['ckanapi', 'load', thing, '--worker']
        + a('--config')
        + a('--ckan-user')
        + a('--remote')
        + a('--apikey')
        + b('--create-only')
        + b('--update-only')
        + b('--upload-resources')
        + b('--upload-logo')
        )


def _copy_from_existing_for_update(obj, existing, thing):
    """
    modifies obj dict in place, copying values from existing.

    the id is alwasys copied from existing to make sure update updates
    the correct object.

    users lists for groups and orgs are maintained if not present in obj
    """
    if 'id' in existing:
        obj['id'] = existing['id']

    if thing in ('organizations', 'groups'):
        if 'users' not in obj and 'users' in existing:
            obj['users'] = existing['users']

def _upload_resources(ckan,obj,arguments):
    resources = obj['resources']
    if not arguments['--upload-resources']:
        return
    for resource in resources:
        if resource.get('url_type') != 'upload':
            continue

        f = requests.get(resource['url'],stream=True)
        name = resource['url'].rsplit('/',1)[-1]
        ckan.call_action('resource_patch',
            {'id':resource['id']},
            files={'upload':(name, f.raw)})


def _upload_logo(ckan,obj):
    for key in obj.keys():
        if isinstance(obj[key],(dict,list)):
            obj.pop(key)                            #dict/list objects can't be encoded
    if urlparse(obj['image_url']).netloc:                  # logo is an external link
        obj['clear_upload'] = True
        obj['image_upload'] = obj['image_url']
    else:
        f = requests.get(obj['image_display_url'],stream=True)
        name,ext = obj['image_url'].rsplit('.',1)  #reformulate image_url for new site
        new_name = re.sub('[0-9\.-]','',name)
        new_url = new_name+'.'+ext
        obj['image_upload'] = (new_url, f.raw)
    ckan.action.group_update(**obj)



"""
ckanapi
-------

This module a thin wrapper around the CKAN's action API.
"""

from ckanapi.errors import (
    CKANAPIError,
    NotAuthorized,
    NotFound,
    ValidationError,
    SearchQueryError,
    SearchError,
    SearchIndexError,
    ServerIncompatibleError,
    )
from ckanapi.localckan import LocalCKAN
from ckanapi.remoteckan import RemoteCKAN
from ckanapi.testappckan import TestAppCKAN









import subprocess
import time
import os
import atexit
import socket
import requests

from ckanapi import RemoteCKAN, NotFound
try:
    import unittest2 as unittest
except ImportError:
    import unittest
try:
    from subprocess import DEVNULL
except ImportError:
    DEVNULL = open(os.devnull, 'wb')
try:
    from urllib2 import urlopen, URLError
except ImportError:
    from urllib.request import urlopen, URLError
try:
    from cStringIO import StringIO
except ImportError:
    from io import StringIO

TEST_CKAN = 'http://localhost:8901'

NUMBER_THING_CSV = """
Number,Thing
5,sasquach
""".lstrip()

class TestRemoteAction(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        script = os.path.join(os.path.dirname(__file__), 'mock/mock_ckan.py')
        _mock_ckan = subprocess.Popen(['python2', script],
            stdout=DEVNULL, stderr=DEVNULL)
        def kill_child():
            try:
                _mock_ckan.kill()
                _mock_ckan.wait()
            except OSError:
                pass  # alread cleaned up from tearDownClass
        atexit.register(kill_child)
        cls._mock_ckan = _mock_ckan
        while True: # wait for the server to start
            try:
                r = urlopen(TEST_CKAN + '/api/action/site_read')
                if r.getcode() == 200:
                    break
            except URLError as e:
                pass
            time.sleep(0.1)

    def test_good_oldstyle(self):
        ckan = RemoteCKAN(TEST_CKAN)
        self.assertEqual(
            ckan.action.organization_list(),
            ['aa', 'bb', 'cc'])
        ckan.close()

    def test_good(self):
        with RemoteCKAN(TEST_CKAN) as ckan:
            self.assertEqual(
                ckan.action.organization_list(),
                ['aa', 'bb', 'cc'])

    def test_missing(self):
        with RemoteCKAN(TEST_CKAN) as ckan:
            self.assertRaises(
                NotFound,
                ckan.action.organization_show,
                id='qqq')

    def test_default_ua(self):
        with RemoteCKAN(TEST_CKAN) as ckan:
            self.assertTrue(
                ckan.action.test_echo_user_agent().startswith('ckanapi'))

    def test_custom_ua(self):
        ua = 'testckanapibot/1.0 (+https://github.com/ckan/ckanapi)'
        with RemoteCKAN(TEST_CKAN, user_agent=ua) as ckan:
            self.assertEqual(ckan.action.test_echo_user_agent(), ua)

    def test_default_content_type(self):
        with RemoteCKAN(TEST_CKAN) as ckan:
            self.assertEqual(ckan.action.test_echo_content_type(),
                "application/json")

    def test_resource_upload(self):
        with RemoteCKAN(TEST_CKAN) as ckan:
            res = ckan.call_action('test_upload',
                {'option': "42"},
                files={'upload': StringIO(NUMBER_THING_CSV)})
        self.assertEqual(res.get('last_row'), ['5', 'sasquach'])

    def test_resource_upload_extra_param(self):
        with RemoteCKAN(TEST_CKAN) as ckan:
            res = ckan.call_action('test_upload',
                {'option': "42"},
                files={'upload': StringIO(NUMBER_THING_CSV)})
        self.assertEqual(res.get('option'), "42")

    def test_resource_upload_unicode_param(self):
        uname = b't\xc3\xab\xc3\x9ft resource'.decode('utf-8')
        with RemoteCKAN(TEST_CKAN) as ckan:
            res = ckan.call_action('test_upload',
                {'option': uname},
                files={'upload': StringIO(NUMBER_THING_CSV)})
        self.assertEqual(res.get('option'), uname)

    def test_resource_upload_content_type(self):
        with RemoteCKAN(TEST_CKAN) as ckan:
            res = ckan.call_action('test_echo_content_type',
                {'option': "42"},
                files={'upload': StringIO(NUMBER_THING_CSV)})
        self.assertEqual(res.split(';')[0], "multipart/form-data")

    @classmethod
    def tearDownClass(cls):
        cls._mock_ckan.kill()
        cls._mock_ckan.wait()




# Copyright (c) 2011-2016 Godefroid Chapelle and ipdb development team
#
# This file is part of ipdb.
# Redistributable under the revised BSD license
# https://opensource.org/licenses/BSD-3-Clause

from __future__ import print_function
import os
import sys

from decorator import contextmanager

__version__ = '0.13.9'

from IPython import get_ipython
from IPython.core.debugger import BdbQuit_excepthook
from IPython.terminal.ipapp import TerminalIPythonApp
from IPython.terminal.embed import InteractiveShellEmbed
try:
    import configparser
except:
    import ConfigParser as configparser


def _get_debugger_cls():
    shell = get_ipython()
    if shell is None:
        # Not inside IPython
        # Build a terminal app in order to force ipython to load the
        # configuration
        ipapp = TerminalIPythonApp()
        # Avoid output (banner, prints)
        ipapp.interact = False
        ipapp.initialize(["--no-term-title"])
        shell = ipapp.shell
    else:
        # Running inside IPython

        # Detect if embed shell or not and display a message
        if isinstance(shell, InteractiveShellEmbed):
            sys.stderr.write(
                "\nYou are currently into an embedded ipython shell,\n"
                "the configuration will not be loaded.\n\n"
            )

    # Let IPython decide about which debugger class to use
    # This is especially important for tools that fiddle with stdout
    return shell.debugger_cls


def _init_pdb(context=None, commands=[]):
    if context is None:
        context = os.getenv("IPDB_CONTEXT_SIZE", get_context_from_config())
    debugger_cls = _get_debugger_cls()
    try:
        p = debugger_cls(context=context)
    except TypeError:
        p = debugger_cls()
    p.rcLines.extend(commands)
    return p


def wrap_sys_excepthook():
    # make sure we wrap it only once or we would end up with a cycle
    #  BdbQuit_excepthook.excepthook_ori == BdbQuit_excepthook
    if sys.excepthook != BdbQuit_excepthook:
        BdbQuit_excepthook.excepthook_ori = sys.excepthook
        sys.excepthook = BdbQuit_excepthook


def set_trace(frame=None, context=None, cond=True):
    if not cond:
        return
    wrap_sys_excepthook()
    if frame is None:
        frame = sys._getframe().f_back
    p = _init_pdb(context).set_trace(frame)
    if p and hasattr(p, 'shell'):
        p.shell.restore_sys_module_state()


def get_context_from_config():
    try:
        parser = get_config()
        return parser.getint("ipdb", "context")
    except (configparser.NoSectionError, configparser.NoOptionError):
        return 3
    except ValueError:
        value = parser.get("ipdb", "context")
        raise ValueError(
            "In %s,  context value [%s] cannot be converted into an integer."
            % (parser.filepath, value)
        )


class ConfigFile(object):
    """
    Filehandle wrapper that adds a "[ipdb]" section to the start of a config
    file so that users don't actually have to manually add a [ipdb] section.
    Works with configparser versions from both Python 2 and 3
    """

    def __init__(self, filepath):
        self.first = True
        with open(filepath) as f:
            self.lines = f.readlines()

    # Python 2.7 (Older dot versions)
    def readline(self):
        try:
            return self.__next__()
        except StopIteration:
            return ''

    # Python 2.7 (Newer dot versions)
    def next(self):
        return self.__next__()

    # Python 3
    def __iter__(self):
        return self

    def __next__(self):
        if self.first:
            self.first = False
            return "[ipdb]\n"
        if self.lines:
            return self.lines.pop(0)
        raise StopIteration


def get_config():
    """
    Get ipdb config file settings.
    All available config files are read.  If settings are in multiple configs,
    the last value encountered wins.  Values specified on the command-line take
    precedence over all config file settings.
    Returns: A ConfigParser object.
    """
    parser = configparser.ConfigParser()

    filepaths = []

    # Low priority goes first in the list
    for cfg_file in ("setup.cfg", ".ipdb", "pyproject.toml"):
        cwd_filepath = os.path.join(os.getcwd(), cfg_file)
        if os.path.isfile(cwd_filepath):
            filepaths.append(cwd_filepath)

    # Medium priority (whenever user wants to set a specific path to config file)
    home = os.getenv("HOME")
    if home:
        default_filepath = os.path.join(home, ".ipdb")
        if os.path.isfile(default_filepath):
            filepaths.append(default_filepath)

    # High priority (default files)
    env_filepath = os.getenv("IPDB_CONFIG")
    if env_filepath and os.path.isfile(env_filepath):
        filepaths.append(env_filepath)

    if filepaths:
        # Python 3 has parser.read_file(iterator) while Python2 has
        # parser.readfp(obj_with_readline)
        try:
            read_func = parser.read_file
        except AttributeError:
            read_func = parser.readfp
        for filepath in filepaths:
            parser.filepath = filepath
            # Users are expected to put an [ipdb] section
            # only if they use setup.cfg
            if filepath.endswith('setup.cfg'):
                with open(filepath) as f:
                    parser.remove_section("ipdb")
                    read_func(f)
            # To use on pyproject.toml, put [tool.ipdb] section
            elif filepath.endswith('pyproject.toml'):
                import toml
                toml_file = toml.load(filepath)
                if "tool" in toml_file and "ipdb" in toml_file["tool"]:
                    if not parser.has_section("ipdb"):
                        parser.add_section("ipdb")
                    for key, value in toml_file["tool"]["ipdb"].items():
                        parser.set("ipdb", key, str(value))
            else:
                read_func(ConfigFile(filepath))
    return parser


def post_mortem(tb=None):
    wrap_sys_excepthook()
    p = _init_pdb()
    p.reset()
    if tb is None:
        # sys.exc_info() returns (type, value, traceback) if an exception is
        # being handled, otherwise it returns None
        tb = sys.exc_info()[2]
    if tb:
        p.interaction(None, tb)


def pm():
    post_mortem(sys.last_traceback)


def run(statement, globals=None, locals=None):
    _init_pdb().run(statement, globals, locals)


def runcall(*args, **kwargs):
    return _init_pdb().runcall(*args, **kwargs)


def runeval(expression, globals=None, locals=None):
    return _init_pdb().runeval(expression, globals, locals)


@contextmanager
def launch_ipdb_on_exception():
    try:
        yield
    except Exception:
        e, m, tb = sys.exc_info()
        print(m.__repr__(), file=sys.stderr)
        post_mortem(tb)
    finally:
        pass


# iex is a concise alias
iex = launch_ipdb_on_exception()


_usage = """\
usage: python -m ipdb [-m] [-c command] ... pyfile [arg] ...

Debug the Python program given by pyfile.

Initial commands are read from .pdbrc files in your home directory
and in the current directory, if they exist.  Commands supplied with
-c are executed after commands from .pdbrc files.

To let the script run until an exception occurs, use "-c continue".
To let the script run up to a given line X in the debugged file, use
"-c 'until X'"

Option -m is available only in Python 3.7 and later.

ipdb version %s.""" % __version__


def main():
    import traceback
    import sys
    import getopt

    try:
        from pdb import Restart
    except ImportError:
        class Restart(Exception):
            pass

    if sys.version_info >= (3, 7):
        opts, args = getopt.getopt(sys.argv[1:], 'mhc:', ['help', 'command='])
    else:
        opts, args = getopt.getopt(sys.argv[1:], 'hc:', ['help', 'command='])

    commands = []
    run_as_module = False
    for opt, optarg in opts:
        if opt in ['-h', '--help']:
            print(_usage)
            sys.exit()
        elif opt in ['-c', '--command']:
            commands.append(optarg)
        elif opt in ['-m']:
            run_as_module = True

    if not args:
        print(_usage)
        sys.exit(2)

    mainpyfile = args[0]     # Get script filename
    if not run_as_module and not os.path.exists(mainpyfile):
        print('Error:', mainpyfile, 'does not exist')
        sys.exit(1)

    sys.argv = args     # Hide "pdb.py" from argument list

    # Replace pdb's dir with script's dir in front of module search path.
    if not run_as_module:
        sys.path[0] = os.path.dirname(mainpyfile)

    # Note on saving/restoring sys.argv: it's a good idea when sys.argv was
    # modified by the script being debugged. It's a bad idea when it was
    # changed by the user from the command line. There is a "restart" command
    # which allows explicit specification of command line arguments.
    pdb = _init_pdb(commands=commands)
    while 1:
        try:
            if run_as_module:
                pdb._runmodule(mainpyfile)
            else:
                pdb._runscript(mainpyfile)
            if pdb._user_requested_quit:
                break
            print("The program finished and will be restarted")
        except Restart:
            print("Restarting", mainpyfile, "with arguments:")
            print("\t" + " ".join(sys.argv[1:]))
        except SystemExit:
            # In most cases SystemExit does not warrant a post-mortem session.
            print("The program exited via sys.exit(). Exit status: ", end='')
            print(sys.exc_info()[1])
        except:
            traceback.print_exc()
            print("Uncaught exception. Entering post mortem debugging")
            print("Running 'cont' or 'step' will restart the program")
            t = sys.exc_info()[2]
            pdb.interaction(None, t)
            print("Post mortem debugger finished. The " + mainpyfile +
                  " will be restarted")


if __name__ == '__main__':
    main()



import sys
import click
from ckanapi import RemoteCKAN
from dpckan.validations import run_resource_validations
from dpckan.functions import (
                              load_complete_datapackage,
                              update_datapackage_json_resource, 
                              resource_create,
                              get_ckan_dataset_resources_ids,
                              get_ckan_datapackage_resource_id,
                              dataset_patch_resources_ids,
                              dataset_path,
                              )

def create(ckan_host, ckan_key, datapackage, datastore, exit_code, resource_name):
  local_datapackage = load_complete_datapackage(datapackage)
  dataset = dataset_path(ckan_host, local_datapackage)
  click.echo(f"Creating resource on {dataset}")
  ckan_instance = RemoteCKAN(ckan_host, apikey = ckan_key)
  try:
    run_resource_validations(ckan_instance, local_datapackage)
    resource_ckan = resource_create(
                                    ckan_instance,
                                    local_datapackage.name,
                                    local_datapackage.get_resource(resource_name),
                                    datastore,
                                    )
    # Update dataset with new resource id
    local_datapackage['resources_ids'] = get_ckan_dataset_resources_ids(ckan_instance, local_datapackage)
    local_datapackage['resources_ids'][resource_name] = resource_ckan['id']
    ckan_datapackage_resource_id = get_ckan_datapackage_resource_id(ckan_instance, local_datapackage.name)
    dataset_patch_resources_ids(ckan_instance, local_datapackage)
    update_datapackage_json_resource(ckan_instance, local_datapackage, ckan_datapackage_resource_id)
    click.echo(f"Resource {resource_name} created on {dataset}")
  except Exception:
    print(f"Error creating {resource_name} resource on {dataset}")
    if exit_code == True:
      sys.exit(1)

@click.command(name='create')
@click.pass_context
def create_cli(ctx):
  """
  Create dataset resource in a CKAN instance.
  """
  create(
         ctx.obj['CKAN_HOST'],
         ctx.obj['CKAN_KEY'],
         ctx.obj['DATAPACKAGE'],
         ctx.obj['DATASTORE'],
         ctx.obj['EXIT_CODE'],
         ctx.obj['RESOURCE_NAME'],
        )



class ServerIncompatibleError(Exception):
    """
    The error raised from RemoteCKAN.call_action when the API doesn't behave
    like a CKAN API.
    """


class CKANAPIError(Exception):
    """
    The error raised from RemoteCKAN.call_action when no other error
    is recognized.

    If importing CKAN source fails then new versions of NotAuthorized,
    ValidationError, NotFound, SearchQueryError, SearchError and
    SearchIndexError are created as subclasses of this class so that they
    provide a helpful str() for tracebacks.
    """

    def __init__(self, extra_msg=None):
        self.extra_msg = extra_msg

    def __str__(self):
        return self.extra_msg


class CLIError(Exception):
    pass


try:
    import ckan

except ImportError:
    # Implement the minimum to be compatible with existing errors
    # without requiring CKAN

    class NotAuthorized(CKANAPIError):
        pass

    class ValidationError(CKANAPIError):
        def __init__(self, error_dict):
            self.error_dict = error_dict
        def __str__(self):
            return repr(self.error_dict)

    class NotFound(CKANAPIError):
        def __init__(self, extra_msg=None):
            self.extra_msg = extra_msg
        def __str__(self):
            return self.extra_msg

    class SearchQueryError(CKANAPIError):
        pass

    class SearchError(CKANAPIError):
        pass

    class SearchIndexError(CKANAPIError):
        pass

else:
    # import ckan worked, so these must not fail
    from ckan.logic import (NotAuthorized, NotFound, ValidationError)
    from ckan.lib.search import (SearchQueryError, SearchError,
                                 SearchIndexError)




class Plugin:
    """Plugin representation

    API      | Usage
    -------- | --------
    Public   | `from frictionless import Plugin`

    It's an interface for writing Frictionless plugins.
    You can implement one or more methods to hook into Frictionless system.

    """

    code = "plugin"
    status = "stable"

    def create_candidates(self, candidates):
        """Create candidates

        Returns:
            dict[]: an ordered by priority list of type descriptors for type detection
        """
        pass

    def create_check(self, name, *, descriptor=None):
        """Create check

        Parameters:
            name (str): check name
            descriptor (dict): check descriptor

        Returns:
            Check: check
        """
        pass

    def create_control(self, file, *, descriptor):
        """Create control

        Parameters:
            file (File): control file
            descriptor (dict): control descriptor

        Returns:
            Control: control
        """
        pass

    def create_dialect(self, file, *, descriptor):
        """Create dialect

        Parameters:
            file (File): dialect file
            descriptor (dict): dialect descriptor

        Returns:
            Dialect: dialect
        """
        pass

    def create_error(self, descriptor):
        """Create error

        Parameters:
            descriptor (dict): error descriptor

        Returns:
            Error: error
        """
        pass

    def create_file(self, source, **options):
        """Create file

        Parameters:
            source (any): file source
            options (dict): file options

        Returns:
            File: file
        """
        pass

    def create_loader(self, file):
        """Create loader

        Parameters:
            file (File): loader file

        Returns:
            Loader: loader
        """
        pass

    def create_parser(self, file):
        """Create parser

        Parameters:
            file (File): parser file

        Returns:
            Parser: parser
        """
        pass

    def create_server(self, name):
        """Create server

        Parameters:
            name (str): server name

        Returns:
            Server: server
        """
        pass

    def create_step(self, descriptor):
        """Create step

        Parameters:
            descriptor (dict): step descriptor

        Returns:
            Step: step
        """
        pass

    def create_storage(self, name, source, **options):
        """Create storage

        Parameters:
            name (str): storage name
            options (str): storage options

        Returns:
            Storage: storage
        """
        pass

    def create_type(self, field):
        """Create type

        Parameters:
            field (Field): corresponding field

        Returns:
            Type: type
        """
        pass



"""
implementation of delete cli command
"""

import sys
import gzip
import json
from datetime import datetime
from itertools import chain
import re
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from ckanapi.errors import (NotFound, NotAuthorized, ValidationError,
    SearchIndexError)
from ckanapi.cli import workers
from ckanapi.cli.utils import completion_stats, compact_json, quiet_int_pipe

try:
    unicode
except NameError:
    unicode = str

def delete_things(ckan, thing, arguments,
        worker_pool=None, stdin=None, stdout=None, stderr=None):
    """
    delete datasets, groups, orgs, users etc,

    The parent process creates a pool of worker processes and hands
    out json lines to each worker as they finish a task. Status of
    last record completed and records being processed is displayed
    on stderr.
    """
    if worker_pool is None:
        worker_pool = workers.worker_pool
    if stdin is None:
        stdin = getattr(sys.stdin, 'buffer', sys.stdin)
    if stdout is None:
        stdout = getattr(sys.__stdout__, 'buffer', sys.__stdout__)
    if stderr is None:
        stderr = getattr(sys.stderr, 'buffer', sys.stderr)

    if arguments['--worker']:
        return delete_things_worker(ckan, thing, arguments)

    log = None
    if arguments['--log']:
        log = open(arguments['--log'], 'a')

    jsonl_input = stdin
    if arguments['--input']:
        jsonl_input = open(arguments['--input'], 'rb')
    if arguments['--gzip']:
        jsonl_input = gzip.GzipFile(fileobj=jsonl_input)

    def name_reader():
        """
        handle start-record and max-records options and extract all
        ids or names from each line (e.g. package_search, package_show
        or package_list output)
        record numbers here correspond to names/ids extracted not lines
        """
        start_record = int(arguments['--start-record'])
        max_records = arguments['--max-records']
        if max_records is not None:
            max_records = int(max_records)

        for num, name in enumerate(chain.from_iterable(
                extract_ids_or_names(line) for line in jsonl_input), 1):
            if num < start_record:
                continue
            if max_records is not None and num >= start_record + max_records:
                break
            yield num, compact_json(name)

    cmd = _worker_command_line(thing, arguments)
    processes = int(arguments['--processes'])
    if hasattr(ckan, 'parallel_limit'):
        # add your sites to ckanapi.remoteckan.MY_SITES instead of removing
        processes = min(processes, ckan.parallel_limit)
    stats = completion_stats(processes)
    if not arguments['ID_OR_NAME']:
        pool = worker_pool(cmd, processes, name_reader())
    else:
        pool = worker_pool(cmd, processes, enumerate(
            (compact_json(n) + b'\n' for n in arguments['ID_OR_NAME']), 1))

    with quiet_int_pipe() as errors:
        for job_ids, finished, result in pool:
            if not result:
                # child exited with traceback
                return 1
            timestamp, error, response = json.loads(
                result.decode('utf-8'))

            if not arguments['--quiet']:
                stderr.write(('%s %s %s %s %s\n' % (
                    finished,
                    job_ids,
                    next(stats),
                    error,
                    compact_json(response).decode('utf-8') if response else ''
                    )).encode('utf-8'))

            if log:
                log.write(compact_json([
                    timestamp,
                    finished,
                    error,
                    response,
                    ]) + b'\n')
                log.flush()
    if 'pipe' in errors:
        return 1
    if 'interrupt' in errors:
        return 2


def extract_ids_or_names(line):
    """
    Be generous in what we accept:

    line may contain
    1. a JSON object with an "id" or "name" value (e.g. package_show result)
    2. a JSON object with a "results" value containing a list
       of objects with "id" values (e.g. package_search result)
    3. a JSON string id or name value
    4. a JSON list of string id or name values (e.g. package_list)
    5. a simple string id or name value

    Returns a list of ids or names found in line
    """
    try:
        j = json.loads(line)
    except ValueError:
        return [line.strip()]  # 5
    if isinstance(j, list) and all(
            isinstance(e, unicode) for e in j):
        return j  # 4
    elif isinstance(j, unicode):
        return [j]  # 3
    elif isinstance(j, dict):
        if 'id' in j and isinstance(j['id'], unicode):
            return [j['id']]  # 1
        if 'name' in j and isinstance(j['name'], unicode):
            return [j['name']]  # 1 again
        if 'results' in j and isinstance(j['results'], list):
            out = []
            for r in j['results']:
                if (not isinstance(r, dict) or 'id' not in r or
                        not isinstance(r['id'], unicode)):
                    break
                out.append(r['id'])
            else:
                return out

    # 5 again (e.g. "true" or "null" or something stranger)
    return [line.strip()]


def delete_things_worker(ckan, thing, arguments,
        stdin=None, stdout=None):
    """
    a process that accepts lines of json on stdin which is parsed and
    passed to the {thing}_delete actions.  it produces lines of json
    which are the responses from each action call.
    """
    if stdin is None:
        stdin = getattr(sys.stdin, 'buffer', sys.stdin)
        # hack so that pdb can be used in extension/ckan
        # code called by this worker
        try:
            sys.stdin = open('/dev/tty', 'rb')
        except IOError:
            pass
    if stdout is None:
        stdout = getattr(sys.__stdout__, 'buffer', sys.__stdout__)
        # hack so that "print debugging" can work in extension/ckan
        # code called by this worker
        sys.stdout = sys.stderr

    thing_delete = {
        'datasets': 'package_delete',
        'groups': 'group_delete',
        'organizations': 'organization_delete',
        'users': 'user_delete',
        'related': 'related_delete',
        }[thing]

    def reply(error, response):
        """
        format messages to be sent back to parent process
        """
        stdout.write(compact_json([
            datetime.now().isoformat(),
            error,
            response]) + b'\n')
        stdout.flush()

    for line in iter(stdin.readline, b''):
        try:
            name = json.loads(line.decode('utf-8'))
        except UnicodeDecodeError as e:
            reply('UnicodeDecodeError', unicode(e))
            continue

        try:
            ckan.call_action(thing_delete, {'id': name})
        except NotAuthorized as e:
            reply('NotAuthorized', unicode(e))
        except NotFound:
            reply('NotFound', name)
        else:
            reply(None, name)

def _worker_command_line(thing, arguments):
    """
    Create a worker command line suitable for Popen with only the
    options the worker process requires
    """
    def a(name):
        "options with values"
        return [name, arguments[name]] * (arguments[name] is not None)
    return (
        ['ckanapi', 'delete', thing, '--worker']
        + a('--config')
        + a('--ckan-user')
        + a('--remote')
        + a('--apikey')
        )



# coding=utf-8
import json

try:
    json_parse_exception = json.decoder.JSONDecodeError
except AttributeError:  # Testing against Python 2
    json_parse_exception = ValueError


resource_mapping = {
    'bytes': 'size',
    'mediatype': 'mimetype',
    'path': 'url'
}

package_mapping = {
    'description': 'notes',
    'homepage': 'url',
}

# Any key not in this list is passed as is inside "extras".
# Further processing will happen for possible matchings, e.g.
# contributor <=> author
ckan_package_keys = [
    'author',
    'author_email',
    'creator_user_id',
    'groups',
    'id',
    'license_id',
    'license_title',
    'license_url',
    'maintainer',
    'maintainer_email',
    'metadata_created',
    'metadata_modified',
    'name',
    'notes',
    'owner_org',
    'private',
    'relationships_as_object',
    'relationships_as_subject',
    'revision_id',
    'resources',
    'state',
    'tags',
    'tracking_summary',
    'title',
    'type',
    'url',
    'version'
]

frictionless_package_keys_to_exclude = [
    'extras'
]


def resource(fddict):
    '''Convert a Frictionless resource to a CKAN resource.

    # TODO: (the following is inaccurate)

    1. Map keys from Frictionless to CKAN (and reformat if needed).
    2. Apply special formatting (if any) for key fields e.g. slugify.
    '''
    resource = dict(fddict)

    # Remap differences from Frictionless to CKAN resource
    for key, value in resource_mapping.items():
        if key in resource:
            resource[value] = resource[key]
            del resource[key]

    return resource


def package(fddict):
    '''Convert a Frictionless package to a CKAN package (dataset).

    # TODO: (the following is inaccurate)

    1. Map keys from Frictionless to CKAN (and reformat if needed).
    2. Apply special formatting (if any) for key fields.
    3. Copy extras across inside the "extras" key.
    '''
    outdict = dict(fddict)

    # Map data package keys
    for key, value in package_mapping.items():
        if key in fddict:
            outdict[value] = fddict[key]
            del outdict[key]

    # map resources inside dataset
    if 'resources' in fddict:
        outdict['resources'] = [resource(res) for res in fddict['resources']]

    if 'licenses' in outdict and outdict['licenses']:
        outdict['license_id'] = outdict['licenses'][0].get('name')
        outdict['license_title'] = outdict['licenses'][0].get('title')
        outdict['license_url'] = outdict['licenses'][0].get('path')
        # remove it so it won't get put in extras
        if len(outdict['licenses']) == 1:
            outdict.pop('licenses', None)

    if outdict.get('contributors'):
        for c in outdict['contributors']:
            if c.get('role') in [None, 'author']:
                outdict['author'] = c.get('title')
                outdict['author_email'] = c.get('email')
                break

        for c in outdict['contributors']:
            if c.get('role') == 'maintainer':
                outdict['maintainer'] = c.get('title')
                outdict['maintainer_email'] = c.get('email')
                break

        # we remove contributors where we have extracted everything into
        # ckan core that way it won't end up in extras
        # this helps ensure that round tripping with ckan is good
        # when have we extracted everything?
        # if contributors has length 1 and role in author or maintainer
        # or contributors == 2 and no of authors and maintainer types <= 1
        if (
            (len(outdict.get('contributors')) == 1 and
                outdict['contributors'][0].get('role') in [None, 'author',
                    'maintainer'])
            or
            (len(outdict.get('contributors')) == 2 and
                [c.get('role') for c in outdict['contributors']]
                not in (
                    [None, None],
                    ['maintainer', 'maintainer'],
                    ['author', 'author']))
                    ):
            outdict.pop('contributors', None)

    if outdict.get('keywords'):
        outdict['tags'] = [
            {'name': keyword} for keyword in outdict['keywords']
        ]
        del outdict['keywords']

    final_dict = dict(outdict)
    for key, value in outdict.items():
        if (
            key not in ckan_package_keys and
            key not in frictionless_package_keys_to_exclude
        ):
            if isinstance(value, (dict, list)):
                value = json.dumps(value)
            if not final_dict.get('extras'):
                final_dict['extras'] = []
            final_dict['extras'].append(
                {'key': key, 'value': value}
            )
            del final_dict[key]
    outdict = dict(final_dict)

    return outdict



import sys
from ckan.lib.cli import CkanCommand

from ckanapi.cli import main

class _Options(object):
    pass

class _DelegateParsing(object):
    usage = main.__doc__

    def parse_args(self, args):
        assert sys.argv[1:3] == ['--plugin=ckanapi', 'ckanapi'], sys.argv
        del sys.argv[1:3]
        arguments = main.parse_arguments()
        cfg = arguments['--config']
        options = _Options()
        options.config = cfg if cfg is not None else './development.ini'
        return options, []

class CKANAPICommand(CkanCommand):
    summary = main.__doc__.split('\n')[0]
    usage = main.__doc__
    parser = _DelegateParsing()

    def command(self):
        self._load_config()

        return main.main(running_with_paster=True)



from ckanapi.cli.action import action
from ckanapi.errors import CLIError
try:
    import unittest2 as unittest
except ImportError:
    import unittest

from io import BytesIO


class MockCKAN(object):
    def __init__(self, expected_name, expected_args, response,
            expected_files=None):
        self._expected_name = expected_name
        self._expected_args = expected_args
        self._expected_files = expected_files or {}
        self._response = response

    def call_action(self, name, args, context=None, apikey=None, files=None):
        if name != self._expected_name:
            return ["wrong name", name, self._expected_name]
        if args != self._expected_args:
            return ["wrong args", args, self._expected_args]
        files = dict((f, v.name) for f,v in files.items())
        if files != self._expected_files:
            return ["wrong files", files, self._expected_files]
        return self._response


class TestCLIAction(unittest.TestCase):
    def test_pretty(self):
        ckan = MockCKAN('shake_it', {'who': 'me'}, {"oh": ["right", "on"]})
        rval = action(ckan, {
            'ACTION_NAME': 'shake_it',
            'KEY=STRING': ['who=me'],
            '--output-json': False,
            '--output-jsonl': False,
            '--input-json': False,
            '--input': None,
            })
        self.assertEqual(b''.join(rval), b"""
{
  "oh": [
    "right",
    "on"
  ]
}
""".lstrip())

    def test_compact(self):
        ckan = MockCKAN('shake_it', {'who': 'me'}, ["right", "on"])
        rval = action(ckan, {
            'ACTION_NAME': 'shake_it',
            'KEY=STRING': ['who=me'],
            '--output-json': True,
            '--output-jsonl': False,
            '--input-json': False,
            '--input': None,
            })
        self.assertEqual(b''.join(rval), b'["right","on"]\n')

    def test_compact_fallback(self):
        ckan = MockCKAN('shake_it', {'who': 'me'}, {"oh": ["right", "on"]})
        rval = action(ckan, {
            'ACTION_NAME': 'shake_it',
            'KEY=STRING': ['who=me'],
            '--output-json': False,
            '--output-jsonl': True,
            '--input-json': False,
            '--input': None,
            })
        self.assertEqual(b''.join(rval), b'{"oh":["right","on"]}\n')

    def test_jsonl(self):
        ckan = MockCKAN('shake_it', {'who': 'me'}, [99,98,97])
        rval = action(ckan, {
            'ACTION_NAME': 'shake_it',
            'KEY=STRING': ['who=me'],
            '--output-json': False,
            '--output-jsonl': True,
            '--input-json': False,
            '--input': None,
            })
        self.assertEqual(b''.join(rval), b'99\n98\n97\n')

    def test_stdin_json(self):
        ckan = MockCKAN('shake_it', {'who': ['just', 'me']}, "yeah")
        rval = action(ckan, {
                'ACTION_NAME': 'shake_it',
                'KEY=STRING': ['who=me'],
                '--output-json': False,
                '--output-jsonl': False,
                '--input-json': True,
                '--input': None,
            },
            stdin=BytesIO(b'{"who":["just","me"]}'),
            )
        self.assertEqual(b''.join(rval), b'"yeah"\n')

    def test_key_json(self):
        ckan = MockCKAN('shake_it', {'who': ['just', 'me']}, "yeah")
        rval = action(ckan, {
            'ACTION_NAME': 'shake_it',
            'KEY=STRING': ['who:["just", "me"]'],
            '--output-json': False,
            '--output-jsonl': False,
            '--input-json': False,
            '--input': None,
            })
        self.assertEqual(b''.join(rval), b'"yeah"\n')

    def test_bad_arg(self):
        ckan = MockCKAN('shake_it', {'who': 'me'}, "yeah")
        rval = action(ckan, {
            'ACTION_NAME': 'shake_it',
            'KEY=STRING': ['who'],
            '--output-json': False,
            '--output-jsonl': False,
            '--input-json': False,
            '--input': None,
            })
        self.assertRaises(CLIError, list, rval)

    def test_bad_key_json(self):
        ckan = MockCKAN('shake_it', {'who': 'me'}, "yeah")
        rval = action(ckan, {
            'ACTION_NAME': 'shake_it',
            'KEY=STRING': ['who:me'],
            '--output-json': False,
            '--output-jsonl': False,
            '--input-json': False,
            '--input': None,
            })
        self.assertRaises(CLIError, list, rval)

    def test_key_string_or_json(self):
        ckan = MockCKAN('shake_it', {'who': 'me=you'}, "yeah")
        rval = action(ckan, {
            'ACTION_NAME': 'shake_it',
            'KEY=STRING': ['who:"me=you"'],
            '--output-json': False,
            '--output-jsonl': False,
            '--input-json': False,
            '--input': None,
            })
        self.assertEqual(b''.join(rval), b'"yeah"\n')

    def test_key_json_or_string(self):
        ckan = MockCKAN('shake_it', {'who': 'me:you'}, "yeah")
        rval = action(ckan, {
            'ACTION_NAME': 'shake_it',
            'KEY=STRING': ['who=me:you'],
            '--output-json': False,
            '--output-jsonl': False,
            '--input-json': False,
            '--input': None,
            })
        self.assertEqual(b''.join(rval), b'"yeah"\n')



import sys
import click
from ckanapi import RemoteCKAN
from dpckan.validations import run_dataset_validations
from dpckan.functions import (
                              delete_dataset,
                              dataset_create,
                              load_complete_datapackage,
                              dataset_path,
                              )

def create(ckan_host, ckan_key, datapackage, datastore, exit_code):
  local_datapackage = load_complete_datapackage(datapackage)
  dataset = dataset_path(ckan_host, local_datapackage)
  click.echo(f"Creating dataset {dataset}")
  ckan_instance = RemoteCKAN(ckan_host, apikey = ckan_key)
  try:
    run_dataset_validations(ckan_instance, local_datapackage)
    dataset_create(ckan_instance, local_datapackage, datastore)
    click.echo(f"Dataset {dataset} created")
  except Exception:
    delete_dataset(ckan_instance, local_datapackage.name)
    print(f"Error during {dataset} creation")
    if exit_code == True:
      sys.exit(1)

@click.command(name='create')
@click.pass_context
def create_cli(ctx):
  """
  Create dataset in a CKAN instance.
  """
  create(
         ctx.obj['CKAN_HOST'],
         ctx.obj['CKAN_KEY'],
         ctx.obj['DATAPACKAGE'],
         ctx.obj['DATASTORE'],
         ctx.obj['EXIT_CODE'],
         )



from frictionless import Package
from dpckan.functions import (
                              load_complete_datapackage,
                              get_ckan_dataset_resources_names,
                              find_dataset_basepath,
                              resource_hash,
                              is_uuid,
                              )
import pytest

source = './dpckan/tests/dataset_test_1/datapackage.json'
package = Package(source)

def test_load_complete_datapackage_dict():
    package = load_complete_datapackage(source)
    assert isinstance(package, dict) == True

def test_load_complete_datapackage_package():
    package = load_complete_datapackage(source)
    assert isinstance(package, Package) == True

def test_get_ckan_dataset_resources_names_list():
  ckan_dataset_resources_ids = {'letters-consonants': '780e2e9a-51f1-4571-95af-bedf05da0dc2',
                                'letters-vowels': '9e4f0dfe-5ecb-482d-a30d-8a9f49f7ac01',
                                'datapackage.json': 'd19de465-16d1-445a-934f-d11f29cadbee',}
  ckan_dataset_resources_names = get_ckan_dataset_resources_names(ckan_dataset_resources_ids)
  assert isinstance(ckan_dataset_resources_names, list) == True

@pytest.mark.parametrize("expect",[('./dpckan/tests/dataset_test_1'),('.')])
def test_find_dataset_basepath(expect):
  if expect == './dpckan/tests/dataset_test_1':
    assert find_dataset_basepath(package) == './dpckan/tests/dataset_test_1'
  else:
    package.basepath = ''
    assert find_dataset_basepath(package) == '.'

@pytest.mark.parametrize(
                          "name,expect",
                          [
                           ('letters-consonants', 'a4a881d0d7070e3357a2ded041c51010'),
                           ('letters', 'b8f894cd1bbbaac091920d7d5bf681ea'),
                           # ('datapackage.json', 'bf39f25e44d331fb44e67fb32b6e795b'),                           
                          ])
def test_resource_hash(name,expect):
  assert resource_hash(package, name) == expect

@pytest.mark.parametrize(
                          "_input,expect",
                          [
                           ('078c9abe-d53d-4e3c-8165-fabbe2d20e0c', True),
                           ('078c9abe-d53d-4e3c-8165-fabbe2d20e0', False),
                           ('078c9abe-d53d-4e3c-8165-fabbe2d20e0d', True),
                           ('ç78c9abe-d53d-4e3c-8165-fabbe2d20e0d', False),
                          ])
def test_is_uuid(_input,expect):
  assert is_uuid(_input) == expect


from ckanapi.cli.dump import dump_things, dump_things_worker
from ckanapi.errors import NotFound
import json
import tempfile
import shutil
from os.path import exists

try:
    import unittest2 as unittest
except ImportError:
    import unittest
from io import BytesIO


class MockCKAN(object):
    def call_action(self, name, data_dict):
        try:
            return {
                'package_list': {
                    None: ['12', '34', 'dp']
                    },
                'package_show': {
                    '12': {
                        'id': '12',
                        'name': 'twelve',
                        'title': "Twelve"},
                    '34': {
                        'id': '34',
                        'name': 'thirtyfour',
                        'title': "Thirty-four"},
                    'dp': {
                        'id': 'dp',
                        'name': 'dp',
                        'title': 'Test for datapackage',
                        'resources':[
                            {'name': 'resource1',
                             'format': 'html',
                             'url':'http://example.com/test-file'}]}
                    },
                'group_show': {
                    'ab': {'title': "ABBA"},
                    },
                'organization_show': {
                    'cd': {'title': "Super Trouper"},
                    },
                }[name][data_dict.get('id')]
        except KeyError:
            raise NotFound()


class TestCLIDump(unittest.TestCase):
    def setUp(self):
        self.ckan = MockCKAN()
        self.stdout = BytesIO()
        self.stderr = BytesIO()

    def test_worker_one(self):
        rval = dump_things_worker(self.ckan, 'datasets', {},
            stdin=BytesIO(b'"34"\n'), stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(error, None)
        self.assertEqual(data["title"], "Thirty-four")

    def test_worker_two(self):
        rval = dump_things_worker(self.ckan, 'datasets', {},
            stdin=BytesIO(b'"12"\n"34"\n'), stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response.count(b'\n'), 2, response)
        self.assertEqual(response[-1:], b'\n')
        r1, r2 = response.split(b'\n', 1)
        timstamp, error, data = json.loads(r1.decode('UTF-8'))
        self.assertEqual(error, None)
        self.assertEqual(data["title"], "Twelve")
        timstamp, error, data = json.loads(r2.decode('UTF-8'))
        self.assertEqual(error, None)
        self.assertEqual(data["title"], "Thirty-four")

    def test_worker_error(self):
        dump_things_worker(self.ckan, 'datasets', {},
            stdin=BytesIO(b'"99"\n'), stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(error, "NotFound")
        self.assertEqual(data, None)

    def test_worker_group(self):
        dump_things_worker(self.ckan, 'groups', {},
            stdin=BytesIO(b'"ab"\n'), stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(error, None)
        self.assertEqual(data, {"title":"ABBA"})

    def test_worker_organization(self):
        dump_things_worker(self.ckan, 'organizations', {},
            stdin=BytesIO(b'"cd"\n'), stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(error, None)
        self.assertEqual(data, {"title":"Super Trouper"})

    def test_parent_dump_all(self):
        dump_things(self.ckan, 'datasets', {
                '--quiet': False,
                '--ckan-user': None,
                '--config': None,
                '--remote': None,
                '--apikey': None,
                '--worker': False,
                '--log': None,
                '--output': None,
                '--datapackages': None,
                '--gzip': False,
                '--all': True,
                '--processes': '1',
                '--get-request': False,
            },
            worker_pool=self._mock_worker_pool,
            stdout=self.stdout,
            stderr=self.stderr)
        self.assertEqual(self.worker_cmd, [
            'ckanapi', 'dump', 'datasets', '--worker',
            'value-here-to-make-docopt-happy'])
        self.assertEqual(self.worker_processes, 1)
        self.assertEqual(self.worker_jobs,
            [(0, b'"12"\n'), (1, b'"34"\n'), (2, b'"dp"\n')])

    def test_parent_parallel_limit(self):
        self.ckan.parallel_limit = 2
        dump_things(self.ckan, 'datasets', {
                '--quiet': False,
                '--ckan-user': None,
                '--config': None,
                '--remote': None,
                '--apikey': None,
                '--worker': False,
                '--log': None,
                '--output': None,
                '--datapackages': None,
                '--gzip': False,
                '--all': False,
                'ID_OR_NAME': ['12'],
                '--processes': '5',
                '--get-request': False,
            },
            worker_pool=self._mock_worker_pool,
            stdout=self.stdout,
            stderr=self.stderr)
        self.assertEqual(self.worker_cmd, [
            'ckanapi', 'dump', 'datasets', '--worker',
            'value-here-to-make-docopt-happy'])
        self.assertEqual(self.worker_processes, 2)

    def test_parent_id_argument(self):
        dump_things(self.ckan, 'groups', {
                '--quiet': False,
                '--ckan-user': None,
                '--config': None,
                '--remote': None,
                '--apikey': None,
                '--worker': False,
                '--log': None,
                '--output': None,
                '--datapackages': None,
                '--gzip': False,
                '--all': False,
                'ID_OR_NAME': ['ab'],
                '--processes': '1',
                '--get-request': False,
            },

            worker_pool=self._mock_worker_pool,
            stdout=self.stdout,
            stderr=self.stderr)
        self.assertEqual(self.worker_cmd, [
            'ckanapi', 'dump', 'groups', '--worker',
            'value-here-to-make-docopt-happy'])
        self.assertEqual(self.worker_processes, 1)
        self.assertEqual(self.worker_jobs, [(0, b'"ab"\n')])

    def test_parent_maintain_order(self):
        dump_things(self.ckan, 'organizations', {
                '--quiet': False,
                '--ckan-user': None,
                '--config': None,
                '--remote': None,
                '--apikey': None,
                '--worker': False,
                '--log': None,
                '--output': None,
                '--datapackages': None,
                '--gzip': False,
                '--all': False,
                'ID_OR_NAME': ['P', 'Q', 'R', 'S'],
                '--processes': '1',
                '--get-request': False,
            },
            worker_pool=self._mock_worker_pool_reversed,
            stdout=self.stdout,
            stderr=self.stderr)
        self.assertEqual(self.worker_cmd, [
            'ckanapi', 'dump', 'organizations', '--worker',
            'value-here-to-make-docopt-happy'])
        self.assertEqual(self.worker_processes, 1)
        self.assertEqual(self.stdout.getvalue(),
            b'{"id":"P"}\n'
            b'{"id":"Q"}\n'
            b'{"id":"R"}\n'
            b'{"id":"S"}\n')

    def test_parent_datapackages(self):
        target = tempfile.mkdtemp()
        try:
            dump_things(self.ckan, 'datasets', {
                    '--quiet': False,
                    '--ckan-user': None,
                    '--config': None,
                    '--remote': None,
                    '--apikey': None,
                    '--worker': False,
                    '--log': None,
                    '--output': None,
                    '--datapackages': target,
                    '--gzip': False,
                    '--all': True,
                    '--processes': '1',
                    '--get-request': False,
                },
                worker_pool=self._mock_worker_pool_with_data,
                stdout=self.stdout,
                stderr=self.stderr)
            assert exists(target + '/twelve/datapackage.json')
            assert exists(target + '/thirtyfour/datapackage.json')
            assert exists(target + '/dp/datapackage.json')
            assert exists(target + '/dp/data/test-file')
        finally:
            shutil.rmtree(target)


    def _mock_worker_pool(self, cmd, processes, job_iter):
        self.worker_cmd = cmd
        self.worker_processes = processes
        self.worker_jobs = list(job_iter)
        for i, j in self.worker_jobs:
            jname = json.loads(j.decode('UTF-8'))
            yield [[], i, json.dumps(['some-date', None, {'id': jname}]
                ).encode('UTF-8') + b'\n']

    def _mock_worker_pool_reversed(self, cmd, processes, job_iter):
        return reversed(list(
            self._mock_worker_pool(cmd, processes, job_iter)))

    def _mock_worker_pool_with_data(self, cmd, processes, job_iter):
        self.worker_cmd = cmd
        self.worker_processes = processes
        self.worker_jobs = list(job_iter)
        for i, j in self.worker_jobs:
            jname = json.loads(j.decode('UTF-8'))
            yield [[], i, json.dumps(['some-date', None,
                self.ckan.call_action('package_show', {'id': jname})]
                ).encode('UTF-8') + b'\n']




import ckanapi
try:
    import unittest2 as unittest
except ImportError:
    import unittest


class TestCallAction(unittest.TestCase):
    def test_local_fail(self):
        try:
            import ckan
        except ImportError:
            raise unittest.SkipTest('ckan not importable')
        self.assertRaises(
            ckanapi.CKANAPIError,
            ckanapi.LocalCKAN('fake').call_action,
            'fake', {}, {}, 'apikey not allowed')

    def test_remote_fail(self):
        self.assertRaises(
            ckanapi.CKANAPIError,
            ckanapi.RemoteCKAN('fake').call_action,
            'fake', {}, 'context not allowed')

    def test_test_fail(self):
        self.assertRaises(
            ckanapi.CKANAPIError,
            ckanapi.TestAppCKAN('fake').call_action,
            'fake', {}, 'context not allowed')



"""
implementation of the action cli command
"""

import sys
import json
from os.path import expanduser

from ckanapi.cli.utils import compact_json, pretty_json
from ckanapi.errors import CLIError


def action(ckan, arguments, stdin=None):
    """
    call an action with KEY=STRING, KEY:JSON or JSON args, yield the result
    """
    if stdin is None:
        stdin = getattr(sys.stdin, 'buffer', sys.stdin)

    file_args = {}
    if arguments['--input-json']:
        action_args = json.loads(stdin.read().decode('utf-8'))
    elif arguments['--input']:
        action_args = json.loads(open(
            expanduser(arguments['--input'])).read().decode('utf-8'))
    else:
        action_args = {}
        for kv in arguments['KEY=STRING']:
            skey, p, svalue = kv.partition('=')
            jkey, p, jvalue = kv.partition(':')
            fkey, p, fvalue = kv.partition('@')
            if len(jkey) > len(skey) < len(fkey):
                action_args[skey] = svalue
            elif len(skey) > len(jkey) < len(fkey):
                try:
                    value = json.loads(jvalue)
                except ValueError:
                    raise CLIError("KEY:JSON argument %r has invalid JSON "
                        "value %r" % (jkey, jvalue))
                action_args[jkey] = value
            elif len(jkey) > len(fkey) < len(skey):
                try:
                    f = open(expanduser(fvalue), 'rb')
                except IOError as e:
                    raise CLIError("Error opening %r: %s" %
                        (expanduser(fvalue), e.args[1]))
                file_args[fkey] = f
            else:
                raise CLIError("argument not in the form KEY=STRING, "
                    "KEY:JSON or KEY@FILE %r" % kv)

    result = ckan.call_action(arguments['ACTION_NAME'], action_args,
        files=file_args)

    if arguments['--output-jsonl']:
        if isinstance(result, list):
            for r in result:
                yield compact_json(r) + b'\n'
        else:
            yield compact_json(result) + b'\n'
    elif arguments['--output-json']:
        yield compact_json(result) + b'\n'
    else:
        yield pretty_json(result) + b'\n'



from ckanapi.cli.load import load_things, load_things_worker
from ckanapi.errors import NotFound, ValidationError, NotAuthorized
import json

try:
    import unittest2 as unittest
except ImportError:
    import unittest
from io import BytesIO

class MockCKAN(object):
    def call_action(self, name, data_dict):
        if name == 'package_show' and data_dict['id'] == 'seekrit':
            raise NotAuthorized('naughty user')
        if name == 'package_create' and data_dict.get('name') == '34':
            raise ValidationError({'name': 'That URL is already in use.'})
        if name == 'organization_update':
            if data_dict['id'] == 'used' and data_dict.get('users') != [
                    'people']:
                raise ValidationError({'users': 'should be unchanged'})
            if data_dict['id'] == 'unused' and data_dict.get('users') != []:
                raise ValidationError({'users': 'should be cleared'})
        try:
            return {
                'package_show': {
                    '12': {'title': "Twelve"},
                    '30ish': {'id': '34', 'title': "Thirty-four"},
                    '34': {'id': '34', 'title': "Thirty-four"},
                    },
                'group_show': {
                    'ab': {'title': "ABBA"},
                    },
                'organization_show': {
                    'cd': {'id': 'cd', 'title': "Super Trouper"},
                    'used': {'users': ['people']},
                    'unused': {'users': ['people']},
                    },
                'package_create': {
                    None: {'name': 'something-new'},
                    },
                'package_update': {
                    '34': {'name': 'something-updated'},
                    },
                'group_update': {
                    'ab': {'name': 'group-updated'},
                    },
                'organization_update': {
                    'cd': {'name': 'org-updated'},
                    'used': {'name': 'users-unchanged'},
                    'unused': {'name': 'users-cleared'},
                    },
                'organization_create': {
                    None: {'name': 'org-created'},
                    },
                }[name][data_dict.get('id')]
        except KeyError:
            raise NotFound()


class TestCLILoad(unittest.TestCase):
    def setUp(self):
        self.ckan = MockCKAN()
        self.stdout = BytesIO()
        self.stderr = BytesIO()

    def test_create_with_no_resources(self):
        load_things_worker(self.ckan, 'datasets', {
                '--create-only': False,
                '--update-only': False,
                '--upload-resources':False,
                },
            stdin=BytesIO(b'{"name": "45","title":"Forty-five"}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'create')
        self.assertEqual(error, None)
        self.assertEqual(data, 'something-new')

    def test_create_with_corrupted_resources(self):
        load_things_worker(self.ckan, 'datasets', {
                '--create-only': False,
                '--update-only': False,
                '--upload-resources':False,
                },
            stdin=BytesIO(b'{"name": "45","title":"Forty-five","resources":[{"id":"123"}]}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'create')
        self.assertEqual(error, None)
        self.assertEqual(data, 'something-new')

    def test_create_with_complete_resources(self):
        load_things_worker(self.ckan, 'datasets', {
                '--create-only': False,
                '--update-only': False,
                '--upload-resources':False,
                },
            stdin=BytesIO(
                 b'{"name": "45","title":"Forty-five",'
                 b'"resources":[{"id":"123","url_type":"","url":"http://example.com"}]}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'create')
        self.assertEqual(error, None)
        self.assertEqual(data, 'something-new')

    def test_create_only(self):
        load_things_worker(self.ckan, 'datasets', {
                '--create-only': True,
                '--update-only': False,
                '--upload-resources':False,
                },
            stdin=BytesIO(b'{"name": "45","title":"Forty-five"}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'create')
        self.assertEqual(error, None)
        self.assertEqual(data, 'something-new')

    def test_create_empty_dict(self):
        load_things_worker(self.ckan, 'datasets', {
                '--create-only': False,
                '--update-only': False,
                '--upload-resources':False,
                },
            stdin=BytesIO(b'{}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'create')
        self.assertEqual(error, None)
        self.assertEqual(data, 'something-new')

    def test_create_bad_option(self):
        load_things_worker(self.ckan, 'datasets', {
                '--create-only': False,
                '--update-only': True,
                },
            stdin=BytesIO(b'{"name": "45","title":"Forty-five"}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'show')
        self.assertEqual(error, 'NotFound')
        self.assertEqual(data, [None, '45'])

    def test_update_with_no_resources(self):
        load_things_worker(self.ckan, 'datasets', {
                '--create-only': False,
                '--update-only': False,
                },
            stdin=BytesIO(b'{"name": "30ish","title":"3.4 times ten"}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'update')
        self.assertEqual(error, None)
        self.assertEqual(data, 'something-updated')

    def test_update_with_corrupted_resources(self):
        load_things_worker(self.ckan, 'datasets', {
                '--create-only': False,
                '--update-only': False,
                '--upload-resources': False,
                },
            stdin=BytesIO(b'{"name": "30ish","title":"3.4 times ten","resources":[{"id":"123"}]}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'update')
        self.assertEqual(error, None)
        self.assertEqual(data, 'something-updated')

    def test_update_with_complete_resources(self):
        load_things_worker(self.ckan, 'datasets', {
                '--create-only': False,
                '--update-only': False,
                '--upload-resources': False,
                },
            stdin=BytesIO(
                 b'{"name": "30ish","title":"3.4 times ten",'
                 b'"resources":[{"id":"123","url_type":"","url":"http://example.com"}]}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'update')
        self.assertEqual(error, None)
        self.assertEqual(data, 'something-updated')

    def test_update_only(self):
        load_things_worker(self.ckan, 'datasets', {
                '--create-only': False,
                '--update-only': True,
                '--upload-resources': False,
                },
            stdin=BytesIO(b'{"name": "34","title":"3.4 times ten"}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'update')
        self.assertEqual(error, None)
        self.assertEqual(data, 'something-updated')

    def test_update_bad_option(self):
        load_things_worker(self.ckan, 'datasets', {
                '--create-only': True,
                '--update-only': False,
                '--upload-resources': False,
                },
            stdin=BytesIO(b'{"name": "34","title":"3.4 times ten"}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'create')
        self.assertEqual(error, 'ValidationError')
        self.assertEqual(data, {'name': 'That URL is already in use.'})

    def test_update_unauthorized(self):
        load_things_worker(self.ckan, 'datasets', {
                '--create-only': False,
                '--update-only': False,
                '--upload-resources': False,
                },
            stdin=BytesIO(b'{"name": "seekrit", "title": "Things"}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'show')
        self.assertEqual(error, 'NotAuthorized')
        self.assertEqual(data, 'naughty user')

    def test_update_group(self):
        load_things_worker(self.ckan, 'groups', {
                '--create-only': False,
                '--update-only': False,
                '--upload-resources': False,
                },
            stdin=BytesIO(b'{"id": "ab","title":"a balloon"}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'update')
        self.assertEqual(error, None)
        self.assertEqual(data, 'group-updated')

    def test_update_organization_two(self):
        load_things_worker(self.ckan, 'organizations', {
                '--create-only': False,
                '--update-only': False,
                '--upload-resources': False,
                },
            stdin=BytesIO(
                b'{"name": "cd", "title": "Go"}\n'
                b'{"name": "ef", "title": "Play"}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response.count(b'\n'), 2, response)
        self.assertEqual(response[-1:], b'\n')
        r1, r2 = response.split(b'\n', 1)
        timstamp, action, error, data = json.loads(r1.decode('UTF-8'))
        self.assertEqual(action, 'update')
        self.assertEqual(error, None)
        self.assertEqual(data, 'org-updated')
        timstamp, action, error, data = json.loads(r2.decode('UTF-8'))
        self.assertEqual(action, 'create')
        self.assertEqual(error, None)
        self.assertEqual(data, 'org-created')

    def test_update_organization_with_users_unchanged(self):
        load_things_worker(self.ckan, 'organizations', {
                '--create-only': False,
                '--update-only': False,
                '--upload-resources': False,
                },
            stdin=BytesIO(b'{"id": "used", "title": "here"}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'update')
        self.assertEqual(error, None)
        self.assertEqual(data, 'users-unchanged')

    def test_update_organization_with_users_cleared(self):
        load_things_worker(self.ckan, 'organizations', {
                '--create-only': False,
                '--update-only': False,
                '--upload-resources': False,
                },
            stdin=BytesIO(b'{"id": "unused", "users": []}\n'),
            stdout=self.stdout)
        response = self.stdout.getvalue()
        self.assertEqual(response[-1:], b'\n')
        timstamp, action, error, data = json.loads(response.decode('UTF-8'))
        self.assertEqual(action, 'update')
        self.assertEqual(error, None)
        self.assertEqual(data, 'users-cleared')

    def test_parent_load_two(self):
        load_things(self.ckan, 'datasets', {
                '--quiet': False,
                '--ckan-user': None,
                '--config': None,
                '--remote': None,
                '--apikey': None,
                '--worker': False,
                '--log': None,
                '--gzip': False,
                '--processes': '1',
                '--input': None,
                '--create-only': False,
                '--update-only': False,
                '--start-record': '1',
                '--max-records': None,
                '--upload-resources': False,
                '--upload-logo':False
            },
            worker_pool=self._mock_worker_pool,
            stdin=BytesIO(
                b'{"name": "cd", "title": "Go"}\n'
                b'{"name": "ef", "title": "Play"}\n'
                ),
            stdout=self.stdout,
            stderr=self.stderr)
        self.assertEqual(self.worker_cmd, [
            'ckanapi', 'load', 'datasets', '--worker'])
        self.assertEqual(self.worker_processes, 1)
        self.assertEqual(self.worker_jobs, [
            (1, b'{"name": "cd", "title": "Go"}\n'),
            (2, b'{"name": "ef", "title": "Play"}\n'),
            ])

    def test_parent_load_start_max(self):
        load_things(self.ckan, 'groups', {
                '--quiet': False,
                '--ckan-user': None,
                '--config': None,
                '--remote': None,
                '--apikey': None,
                '--worker': False,
                '--log': None,
                '--gzip': False,
                '--processes': '1',
                '--input': None,
                '--create-only': False,
                '--update-only': False,
                '--start-record': '2',
                '--max-records': '2',
                '--upload-resources': False,
                '--upload-logo':False,
            },
            worker_pool=self._mock_worker_pool,
            stdin=BytesIO(
                b'{"name": "cd", "title": "Go"}\n'
                b'{"name": "ef", "title": "Play"}\n'
                b'{"name": "gh", "title": "Hotel"}\n'
                b'{"name": "ij", "title": "Ambient"}\n'
                ),
            stdout=self.stdout,
            stderr=self.stderr)
        self.assertEqual(self.worker_cmd, [
            'ckanapi', 'load', 'groups', '--worker'])
        self.assertEqual(self.worker_processes, 1)
        self.assertEqual(self.worker_jobs, [
            (2, b'{"name": "ef", "title": "Play"}\n'),
            (3, b'{"name": "gh", "title": "Hotel"}\n'),
            ])

    def test_parent_parallel_limit(self):
        self.ckan.parallel_limit = 2
        load_things(self.ckan, 'datasets', {
                '--quiet': False,
                '--ckan-user': None,
                '--config': None,
                '--remote': None,
                '--apikey': None,
                '--worker': False,
                '--log': None,
                '--gzip': False,
                '--processes': '5',
                '--input': None,
                '--create-only': False,
                '--update-only': False,
                '--start-record': '1',
                '--max-records': None,
                '--upload-resources': False,
                '--upload-logo':False,
            },
            worker_pool=self._mock_worker_pool,
            stdin=BytesIO(
                b'{"name": "cd", "title": "Go"}\n'
                b'{"name": "ef", "title": "Play"}\n'
                ),
            stdout=self.stdout,
            stderr=self.stderr)
        self.assertEqual(self.worker_cmd, [
            'ckanapi', 'load', 'datasets', '--worker'])
        self.assertEqual(self.worker_processes, 2)

    def _mock_worker_pool(self, cmd, processes, job_iter):
        self.worker_cmd = cmd
        self.worker_processes = processes
        self.worker_jobs = list(job_iter)
        for i, j in self.worker_jobs:
            jname = json.loads(j.decode('UTF-8'))
            yield [[], i, json.dumps(['some-date', None, None, {'id':jname}]
                ).encode('UTF-8') + b'\n']



"""ckanapi command line interface

Usage:
  ckanapi action ACTION_NAME
          [(KEY=STRING | KEY:JSON | KEY@FILE ) ... | -i | -I JSON_INPUT]
          [-j | -J] [[-c CONFIG] [-u USER] | -r SITE_URL [-a APIKEY] [-g]]
  ckanapi load datasets
          [--upload-resources] [-I JSONL_INPUT] [-s START] [-m MAX]
          [-p PROCESSES] [-l LOG_FILE] [-n | -o] [-qwz]
          [[-c CONFIG] [-u USER] | -r SITE_URL [-a APIKEY]]
  ckanapi load (groups | organizations)
          [--upload-logo] [-I JSONL_INPUT] [-s START] [-m MAX]
          [-p PROCESSES] [-l LOG_FILE] [-n | -o] [-qwz]
          [[-c CONFIG] [-u USER] | -r SITE_URL [-a APIKEY]]
  ckanapi load (users | related)
          [-I JSONL_INPUT] [-s START] [-m MAX] [-p PROCESSES] [-l LOG_FILE]
          [-n | -o] [-qwz] [[-c CONFIG] [-u USER] | -r SITE_URL [-a APIKEY]]
  ckanapi dump (datasets | groups | organizations | users | related)
          (ID_OR_NAME ... | --all) ([-O JSONL_OUTPUT] | [-D DIRECTORY])
          [-p PROCESSES] [-qwz]
          [[-c CONFIG] [-u USER] | -r SITE_URL [-a APIKEY] [-g]]
  ckanapi delete (datasets | groups | organizations | users | related)
          (ID_OR_NAME ... | [-I JSONL_INPUT] [-s START] [-m MAX])
          [-p PROCESSES] [-l LOG_FILE] [-qwz]
          [[-c CONFIG] [-u USER] | -r SITE_URL [-a APIKEY]]
  ckanapi (-h | --help)
  ckanapi --version

Options:
  -h --help                 show this screen
  --version                 show version
  -a --apikey=APIKEY        API key to use for remote actions
  --all                     all the things
  -c --config=CONFIG        CKAN configuration file for local actions,
                            defaults to ./development.ini if that file exists
  -g --get-request          use GET instead of POST for API calls
  -i --input-json           read json from stdin to send to action
  -I --input=INPUT          input json/ json lines from file instead of stdin
  -j --output-json          output plain json instead of pretty-printed json
  -J --output-jsonl         output list responses as json lines instead of
                            pretty-printed json
  -D --datapackages=DIR     download resources and output as datapackages
                            in DIR instead of metadata-only json lines
  -l --log=LOG_FILE         append messages generated to LOG_FILE
  -m --max-records=MAX      exit after processing MAX records
  -n --create-only          create new records, don't update existing records
  -o --update-only          update existing records, don't create new records
  -O --output=JSONL_OUTPUT  output to json lines file instead of stdout
  -p --processes=PROCESSES  set the number of worker processes [default: 1]
  -q --quiet                don't display progress messages
  -r --remote=URL           URL of CKAN server for remote actions
  -s --start-record=START   start from record number START, where the first
                            record is number 1 [default: 1]
  -u --ckan-user=USER       perform actions as user with this name, uses the
                            site sysadmin user when not specified
  --upload-logo             upload logo image of a group/organization if the
                            image is stored in the original server, otherwise
                            its image url will be used
  --upload-resources        upload resources of a dataset that were uploaded to
                            server. Resources originally linked by external
                            urls will keep the urls,will not be uploaded
  -w --worker               launch worker process, used internally by load
                            and dump commands
  -z --gzip                 read/write gzipped data
"""

import sys
import os
from docopt import docopt
from pkg_resources import load_entry_point

from ckanapi.version import __version__
from ckanapi.remoteckan import RemoteCKAN
from ckanapi.localckan import LocalCKAN
from ckanapi.errors import CLIError
from ckanapi.cli.load import load_things
from ckanapi.cli.dump import dump_things
from ckanapi.cli.delete import delete_things
from ckanapi.cli.action import action


def parse_arguments():
    # docopt is awesome
    return docopt(__doc__, version=__version__)


def main(running_with_paster=False):
    """
    ckanapi command line entry point
    """
    arguments = parse_arguments()

    if not running_with_paster and not arguments['--remote']:
        return _switch_to_paster(arguments)

    if arguments['--remote']:
        ckan = RemoteCKAN(arguments['--remote'],
            apikey=arguments['--apikey'],
            user_agent="ckanapi-cli/{version} (+{url})".format(
                version=__version__,
                url='https://github.com/open-data/ckanapi'),
            get_only=arguments['--get-request'],
            )
    else:
        ckan = LocalCKAN(username=arguments['--ckan-user'])

    if arguments['action']:
        try:
            for r in action(ckan, arguments):
                sys.stdout.write(r)
            return
        except CLIError as e:
            sys.stderr.write(e.args[0] + '\n')
            return 1

    things = ['datasets', 'groups', 'organizations', 'users', 'related']
    thing = [x for x in things if arguments[x]]
    if (arguments['load'] or arguments['dump'] or arguments['delete']
            ) and arguments['--processes'] != '1' and os.name == 'nt':
        sys.stderr.write(
            "multiple worker processes are not supported on windows\n")
        arguments['--processes'] = '1'

    if arguments['load']:
        return load_things(ckan, thing[0], arguments)

    if arguments['dump']:
        return dump_things(ckan, thing[0], arguments)

    if arguments['delete']:
        return delete_things(ckan, thing[0], arguments)

    assert 0, arguments # we shouldn't be here


def _switch_to_paster(arguments):
    """
    With --config we switch to the paster command version of the cli
    """
    sys.argv[1:1] = ["--plugin=ckanapi", "ckanapi"]
    sys.exit(load_entry_point('PasteScript', 'console_scripts', 'paster')())



"""
implementation of dump cli command
"""

import sys
import gzip
import json
from datetime import datetime
import os
import requests

from ckanapi.errors import (NotFound, NotAuthorized, ValidationError,
    SearchIndexError)
from ckanapi.cli import workers
from ckanapi.cli.utils import completion_stats, compact_json, \
    quiet_int_pipe, pretty_json

DL_CHUNK_SIZE = 100 * 1024
DATAPACKAGE_VERSION = '1.0-beta.10'


def dump_things(ckan, thing, arguments,
        worker_pool=None, stdout=None, stderr=None):
    """
    dump all datasets, groups, orgs or users accessible by the connected user

    The parent process creates a pool of worker processes and hands
    out ids to each worker. Status of last record completed and records
    being processed is displayed on stderr.
    """
    if worker_pool is None:
        worker_pool = workers.worker_pool
    if stdout is None:
        stdout = getattr(sys.__stdout__, 'buffer', sys.__stdout__)
    if stderr is None:
        stderr = getattr(sys.stderr, 'buffer', sys.stderr)

    if arguments['--worker']:
        return dump_things_worker(ckan, thing, arguments)

    log = None
    if arguments['--log']:
        log = open(arguments['--log'], 'a')

    jsonl_output = stdout
    if arguments['--datapackages']:  # TODO: do we want to just divert this to devnull?
        jsonl_output = open(os.devnull, 'wb')
    if arguments['--output']:
        jsonl_output = open(arguments['--output'], 'wb')
    if arguments['--gzip']:
        jsonl_output = gzip.GzipFile(fileobj=jsonl_output)
    if arguments['--all']:
        get_thing_list = {
            'datasets': 'package_list',
            'groups': 'group_list',
            'organizations': 'organization_list',
            'users': 'user_list',
            'related' :'related_list',
            }[thing]
        params = dict(
            all_fields=False,  # for user_list
            )
        names = ckan.call_action(get_thing_list, params)

    else:
        names = arguments['ID_OR_NAME']

    if names and isinstance(names[0], dict):
        names = [rec.get('name',rec.get('id')) for rec in names]

    cmd = _worker_command_line(thing, arguments)
    processes = int(arguments['--processes'])
    if hasattr(ckan, 'parallel_limit'):
        # add your sites to ckanapi.remoteckan.MY_SITES instead of removing
        processes = min(processes, ckan.parallel_limit)
    stats = completion_stats(processes)
    pool = worker_pool(cmd, processes,
        enumerate(compact_json(n) + b'\n' for n in names))

    results = {}
    expecting_number = 0
    with quiet_int_pipe() as errors:
        for job_ids, finished, result in pool:
            if not result:
                # child exited with traceback
                return 1
            timestamp, error, record = json.loads(result.decode('utf-8'))
            results[finished] = record

            if not arguments['--quiet']:
                stderr.write('{0} {1} {2} {3} {4}\n'.format(
                    finished,
                    job_ids,
                    next(stats),
                    error,
                    record.get('name', '') if record else '',
                    ).encode('utf-8'))

            if log:
                log.write(compact_json([
                    timestamp,
                    finished,
                    error,
                    record.get('name', '') if record else None,
                    ]) + b'\n')

            datapackages_path = arguments['--datapackages']
            if datapackages_path:
                create_datapackage(record, datapackages_path, stderr)

            # keep the output in the same order as names
            while expecting_number in results:
                record = results.pop(expecting_number)
                if record:
                    # sort keys so we can diff output
                    jsonl_output.write(compact_json(record,
                        sort_keys=True) + b'\n')
                expecting_number += 1
    if 'pipe' in errors:
        return 1
    if 'interrupt' in errors:
        return 2


def dump_things_worker(ckan, thing, arguments,
        stdin=None, stdout=None):
    """
    a process that accepts names on stdin which are
    passed to the {thing}_show actions.  it produces lines of json
    which are the responses from each action call.
    """
    if stdin is None:
        stdin = getattr(sys.stdin, 'buffer', sys.stdin)
        # hack so that pdb can be used in extension/ckan
        # code called by this worker
        try:
            sys.stdin = open('/dev/tty', 'rb')
        except IOError:
            pass
    if stdout is None:
        stdout = getattr(sys.__stdout__, 'buffer', sys.__stdout__)
        # hack so that "print debugging" can work in extension/ckan
        # code called by this worker
        sys.stdout = sys.stderr

    thing_show = {
        'datasets': 'package_show',
        'groups': 'group_show',
        'organizations': 'organization_show',
        'users': 'user_show',
        'related':'related_show'
        }[thing]

    def reply(error, record=None):
        """
        format messages to be sent back to parent process
        """
        stdout.write(compact_json([
            datetime.now().isoformat(),
            error,
            record]) + b'\n')
        stdout.flush()

    for line in iter(stdin.readline, b''):
        try:
            name = json.loads(line.decode('utf-8'))
        except UnicodeDecodeError as e:
            reply('UnicodeDecodeError')
            continue

        try:
            obj = ckan.call_action(thing_show, {'id': name,
                'include_datasets': False,
                'include_password_hash': True,
                })
            reply(None, obj)
        except NotFound:
            reply('NotFound')
        except NotAuthorized:
            reply('NotAuthorized')


def create_datapackage(record, base_path, stderr):
    # TODO: how are we going to handle which resources to
    # leave alone? They're very inconsistent in some instances
    # And I can't imagine anyone wants to download a copy
    # of, for example, the API base endpoint
    resource_formats_to_ignore = ['API', 'api']
    dataset_name = record.get('name', '') if record else ''

    target_dir = '{base_path}/{name}/data'.format(
        base_path=base_path,
        name=dataset_name)

    try:
        os.makedirs(target_dir)
    except Exception as e:
        stderr.write(e.message)

    for resource in record.get('resources', ''):
        if resource.get('name') is not None:
            resource_id = resource['name']
        else:
            resource_id = resource['id']

        resource_filename = os.path.split(resource['url'])[1]

        output = os.path.join(target_dir, resource_filename)

        # Resources can have a free-form address and no internal info, so in those cases
        # we're going to merely save them using the UID. (If they even exist)
        if output.endswith('/'):
            output = os.path.join(output, resource_id)

        resource['path'] = 'data' + output[len(target_dir):]

        try:
            if resource['format'] not in resource_formats_to_ignore:
                r = requests.get(resource['url'], stream=True)
                with open(output, 'wb') as f:
                    for chunk in r.iter_content(chunk_size=DL_CHUNK_SIZE):
                        if chunk: # filter out keep-alive new chunks
                            f.write(chunk)
                            f.flush()
        except requests.ConnectionError:
            stderr.write('URL {url} refused connection. The resource will not be downloaded\n'.format(url=resource['url']))
        except requests.exceptions.RequestException as e:
            stderr.write(e.message)
            stderr.write('\n')

    json_output_name = '{base_path}/{dataset_name}/datapackage.json'.format(
        base_path=base_path, dataset_name=dataset_name)
    with open(json_output_name, 'wb') as out:
        out.write(pretty_json(dict(record, version=DATAPACKAGE_VERSION)))


def _worker_command_line(thing, arguments):
    """
    Create a worker command line suitable for Popen with only the
    options the worker process requires
    """
    def a(name):
        "options with values"
        return [name, arguments[name]] * (arguments[name] is not None)
    def b(name):
        "boolean options"
        return [name] * bool(arguments[name])
    return (
        ['ckanapi', 'dump', thing, '--worker']
        + a('--config')
        + a('--ckan-user')
        + a('--remote')
        + a('--apikey')
        + b('--get-request')
        + ['value-here-to-make-docopt-happy']
        )



# Copyright (c) 2007-2016 Godefroid Chapelle and ipdb development team
#
# This file is part of ipdb.
# Redistributable under the revised BSD license
# https://opensource.org/licenses/BSD-3-Clause

from ipdb.__main__ import set_trace, post_mortem, pm, run, iex        # noqa
from ipdb.__main__ import runcall, runeval, launch_ipdb_on_exception  # noqa

from ipdb.stdout import sset_trace, spost_mortem, spm                 # noqa
from ipdb.stdout import slaunch_ipdb_on_exception                     # noqa



import json
import sys
import click
from frictionless import validate
from dpckan.functions import (
                              is_dataset_published,
                              dataset_path,
                              )

def run_dataset_validations(ckan_instance, package):
  """
    Run validations before dataset publication
  """
  is_host_valid(ckan_instance)
  is_dataset_published_check(ckan_instance, package)
  is_datapackage_valid(package)
  is_owner_org_valid(ckan_instance, package)


def run_resource_validations(ckan_instance, package):
  """
    Run validations before dataset publication
  """
  is_host_valid(ckan_instance)
  is_datapackage_valid(package)
  is_owner_org_valid(ckan_instance, package)

def is_host_valid(ckan_instance):
  demo = ckan_instance
  try:
    ckan_exist = demo.action.site_read()
  except:
    print(f'Host {ckan_instance.address} not exist')
    sys.exit(1)

def is_dataset_published_check(ckan_instance, package):
  dataset = dataset_path(ckan_instance.address, package)
  if is_dataset_published(ckan_instance, package.name):
    click.echo(f'Dataset {dataset} already published')
    sys.exit(1)

def is_owner_org_valid(ckan_instance, package):
  dataset = dataset_path(ckan_instance.address, package)
  result = ckan_instance.action.organization_list(id = package.name)
  if package['owner_org'] not in result:
    click.echo(f"{package['owner_org']} not exist in {dataset}")
    sys.exit(1)

def is_datapackage_valid(package):
  """
  Verifica a existência da chave "datapackage_resource_id" no arquivo datapackage.json para o ambiente desejado (homologação ou produção)
  Existência da chave "datapackage_resource_id" significa que pacote já foi publicado no ambiente desejado
  Arquivo datapackage.json deve estar na raiz do dataset
  Trata erro para:
   - datapackage.json inexistente
   - algum erro de sintaxe no arquivo datapackage.json, ou até mesmo arquivo existente mas em branco

  Parameters
  ----------
    env
    homologacao: homologa.cge.mg.gov.br/
    producao: dados.mg.gov.br/


  Returns
  -------
  string
    valor correspondente a chave "datapackage_resource_id" do ambiente solicitado, quando a mesma existir
  None
    quando a chave não existir
  """
  if f"name" in package.keys():
    pass
  else:
    click.echo("Arquivo datapackage.json sem a propriedade `name` obrigatória")
    sys.exit(1)
  if f"resources" in package.keys():
    pass
  else:
    click.echo("Arquivo datapackage.json sem a propriedade `resources` obrigatória")
    sys.exit(1)
  if f"owner_org" in package.keys():
    pass
  else:
    click.echo("Arquivo datapackage.json sem a propriedade `owner_org` obrigatória")
    sys.exit(1)




"""
Code shared by LocalCKAN, RemoteCKAN and TestCKAN
"""

import json

from ckanapi.errors import (CKANAPIError, NotAuthorized, NotFound,
    ValidationError, SearchQueryError, SearchError, SearchIndexError,
    ServerIncompatibleError)

class ActionShortcut(object):
    """
    ActionShortcut(foo).bar(baz=2) <=> foo.call_action('bar', {'baz':2})

    An instance of this class is used as the .action attribute of
    LocalCKAN and RemoteCKAN instances to provide a short way to call
    actions, e.g::

        pkg = demo.action.package_show(id='adur_district_spending')

    instead of::

        pkg = demo.call_action('package_show', {'id':'adur_district_spending'})

    File-like values (objects with a 'read' attribute) are
    sent as file-uploads::

        pkg = demo.action.resource_update(package_id='foo', upload=open(..))

    becomes::

        pkg = demo.call_action('resource_update',
            {'package_id': 'foo'}, files={'upload': open(..)})

    """
    def __init__(self, ckan):
        self._ckan = ckan

    def __getattr__(self, name):
        def action(**kwargs):
            files = {}
            for k, v in kwargs.items():
                if is_file_like(v):
                    files[k] = v
            if files:
                nonfiles = dict((k, v) for k, v in kwargs.items()
                    if k not in files)
                return self._ckan.call_action(name,
                    data_dict=nonfiles,
                    files=files)
            return self._ckan.call_action(name, data_dict=kwargs)
        return action


def is_file_like(v):
    """
    Return True if this object is file-like or is a tuple in a format
    that the requests library would accept for uploading.
    """
    # see http://docs.python-requests.org/en/latest/user/quickstart/#more-complicated-post-requests
    return hasattr(v, 'read') or (
        isinstance(v, tuple) and len(v) >= 2 and hasattr(v[1], 'read'))


def prepare_action(action, data_dict=None, apikey=None, files=None):
    """
    Return action_url, data_json, http_headers
    """
    if not data_dict:
        data_dict = {}
    headers = {}
    if files:
        # when uploading files all parameters must be strings and
        # no nesting is allowed because request is sent as multipart
        items = data_dict.items()
        data_dict = {}
        for (k, v) in items:
            if v is None:
                continue  # assuming missing will work the same as None
            if isinstance(v, (int, float)):
                v = str(v)
            data_dict[k.encode('utf-8')] = v.encode('utf-8')
    else:
        data_dict = json.dumps(data_dict).encode('ascii')
        headers['Content-Type'] = 'application/json'
    if apikey:
        apikey = str(apikey)
        headers['X-CKAN-API-Key'] = apikey
        headers['Authorization'] = apikey
    url = 'api/action/' + action
    return url, data_dict, headers


def reverse_apicontroller_action(url, status, response):
    """
    Make an API call look like a direct action call by reversing the
    exception -> HTTP response translation that ApiController.action does
    """
    try:
        parsed = json.loads(response)
        if parsed.get('success'):
            return parsed['result']
        if hasattr(parsed, 'get'):
            err = parsed.get('error', {})
        else:
            err = {}
    except (AttributeError, ValueError):
        err = {}

    if not isinstance(err, dict):  # possibly a Socrata API.
        raise ServerIncompatibleError(repr([url, status, response]))

    etype = err.get('__type')
    emessage = err.get('message', '').split(': ', 1)[-1]
    if etype == 'Search Query Error':
        # I refuse to eval(emessage), even if it would be more correct
        raise SearchQueryError(emessage)
    elif etype == 'Search Error':
        # I refuse to eval(emessage), even if it would be more correct
        raise SearchError(emessage)
    elif etype == 'Search Index Error':
        raise SearchIndexError(emessage)
    elif etype == 'Validation Error':
        raise ValidationError(err)
    elif etype == 'Not Found Error':
        raise NotFound(emessage)
    elif etype == 'Authorization Error':
        raise NotAuthorized(err)

    # don't recognize the error
    raise CKANAPIError(repr([url, status, response]))



import select
import subprocess

def worker_pool(popen_arg, num_workers, job_iterable,
        stop_when_jobs_done=True, stop_on_keyboard_interrupt=True,
        popen=None):
    """
    Coroutine to manage a pool of workers that accept jobs as single lines
    of input on stdin and produces results as single lines of output.

    popen_arg - parameter to pass to subprocess.Popen when creating workers
    num_workers - maximum number of workers to create
    job_iterable - iterable producing (job id, job string) tuples where
                   job string should include a single trailing newline
    stop_when_jobs_done - True: generator exits when all jobs are done
    stop_on_keyboard_interrupt - True: generator exits on KeyboardIterrupt

    accepted to send(): job iterable or None, when a new job iterable is
    sent it will replace the previous one used for assigning jobs to workers

    This generator blocks until there is a result from one of the workers.

    yields (currently processing job id list, finished job id, job result)
    tuples as jobs are completed, or (None, None, None) when no jobs remain
    to be completed and stop_when_jobs_done is False.

    currently processing job id list will include None if some workers are
    idle.  job result will include trailing newline.

    when no jobs remain to be completed and stop_when_jobs_done is False a
    new job iterable must be sent to this generator with send().
    """
    if popen is None:
        popen = subprocess.Popen

    workers = []
    job_ids = []
    worker_fds = {}
    job_iter = iter(job_iterable)

    def start_job(worker=None):
        """
        assign a job to exiting or newly created worker subprocess.

        returns (job_id, worker) or (None, None) when no more jobs
        """
        job_id, job_str = next(job_iter, (None, None))
        if job_str is None:
            return None, None
        job_str = job_str.rstrip(b'\n') + b'\n'
        if not worker:
            worker = popen(
                popen_arg,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                )
        worker.stdin.write(job_str)
        worker.stdin.flush()
        return (job_id, worker)

    def assign_jobs():
        """
        start as many jobs as possible given maximum/idle workers
        and available jobs
        """
        while None in job_ids:
            wnum = job_ids.index(None)
            job_ids[wnum], w = start_job(workers[wnum])
            if w is None:
                return

        while len(workers) < num_workers:
            job_id, w = start_job()
            if w is None:
                return
            worker_fds[w.stdout] = len(workers)
            workers.append(w)
            job_ids.append(job_id)

    try:
        assign_jobs()
        while True:
            if all(i is None for i in job_ids):
                if stop_when_jobs_done:
                    return
                new_jobs = yield (None, None, None)
                # require new jobs to be submitted
                job_iter = iter(new_jobs)
                assign_jobs()
                continue

            try:
                readable, _, _ = select.select(worker_fds, [], [])
            except select.error as e:
                if e.args[0] == 10038:
                    # XXX: no many-worker support on windows yet
                    readable = list(worker_fds)[:1]
                else:
                    raise
            except KeyboardInterrupt:
                if stop_on_keyboard_interrupt:
                    return
                raise

            fd = readable[0]
            wnum = worker_fds[fd]
            w = workers[wnum]
            result = w.stdout.readline()
            finished = job_ids[wnum]
            job_ids[wnum], _ = start_job(w)

            new_jobs = yield (job_ids, finished, result)
            if new_jobs:
                job_iter = iter(new_jobs)
                assign_jobs()

    finally:
        for w in workers:
            w.stdin.close()



import sys
import click
from ckanapi import RemoteCKAN
from dpckan.functions import (
  load_complete_datapackage,
  is_dataset_published,
)


def diff_resource(ckan_host, ckan_key, datapackage, resource_name):
  """
  Detect changes between datapackage an the created dataset. 

  Parâmetros:

  -------

  ckan_host: string

    host ou ambiente da instância CKAN para a qual se deseja publicar conjunto de dados.
    Exemplo: https://demo.ckan.org/

  ckan_key: string

    Chave CKAN do usuário e ambiente para a qual se deseja publicar conjunto de dados.

  datapackage: string

    Caminho local para arquivo datapackage.json.

  Retorna:

  -------

  A list of (non-expected) differences between the datapackage and the CKAN dataset

  """
  package = load_complete_datapackage(datapackage)

  ckan_instance = RemoteCKAN(ckan_host, apikey=ckan_key)
  if not is_dataset_published(ckan_instance, package.name):
    raise Exception('Conjunto de dados nao existente.')

  ckan_instance = RemoteCKAN(ckan_host, apikey=ckan_key)
  return resource_data_diff(ckan_instance, package, resource_name)


@click.command()
@click.option('--ckan-host', '-H', envvar='CKAN_HOST', required=True,
              help="Ckan host, exemplo: https://demo.ckan.org/")  # -H para respeitar convenção de -h ser help
@click.option('--ckan-key', '-k', envvar='CKAN_KEY', required=True,
              help="Ckan key autorizando o usuário a realizar publicações/atualizações em datasets")
@click.option('--datapackage', '-dp', required=True, default='datapackage.json')
@click.option('--resource-name', '-rn', required=True)
def diff_resource_cli(ckan_host, ckan_key, datapackage, resource_name):
  """
  Detect changes between datapackage resource an a CKAN resource. 

  Parâmetros:

  ----------

  ckan_host: string (não obrigatório caso variável CKAN_HOST esteja cadastrada na máquina ou em arquivo .env)

    host ou ambiente da instância CKAN para a qual se deseja publicar conjunto de dados.
    Exemplo: https://demo.ckan.org/

  ckan_key: string (não obrigatório caso variável CKAN_KEY esteja cadastrada na máquina ou em arquivo .env)

    Chave CKAN do usuário e ambiente para a qual se deseja publicar conjunto de dados.

  datapackage: string (não obrigatório caso comando seja executado no mesmo diretório do arquivo datapackage.json)

    Caminho local para arquivo datapackage.json.

  resource_name: string

    Nome do recurso, presente no arquivo datapackage.json, que será atualizado.

  Retorna:

  -------

  A list of (non-expected) differences between the datapackage and the CKAN dataset
  """

  diffs, oks = diff_resource(ckan_host, ckan_key, datapackage, resource_name)
  if len(diffs) == 0:
    click.echo("There are no differences")
  else:
    click.echo("Differences detected:")
    for diff in diffs:
      click.echo(f" - On field {diff['field_name']}")
      if diff.get('ckan_value'):
        click.echo(f"   - CKAN value: {diff['ckan_value']}")
      if diff.get('datapackage_value'):
        click.echo(f"   - DataPackage value: {diff['datapackage_value']}")
      if diff.get('error'):
        click.echo(f"   - Error: {diff['error']}")
  if len(oks) == 0:
    click.echo("No equal field found")
  else:
    click.echo("Equal fields: {}".format(', '.join(oks)))



# coding=utf-8
import json

try:
    json_parse_exception = json.decoder.JSONDecodeError
except AttributeError:  # Testing against Python 2
    json_parse_exception = ValueError


resource_mapping = {
    'size': 'bytes',
    'mimetype': 'mediatype',
    'url': 'path'
}

resource_keys_to_remove = [
    'position',
    'datastore_active',
    'state'
]


def resource(ckandict):
    '''Convert a CKAN resource to Frictionless Resource.

    1. Remove unneeded keys
    2. Expand extras.
        * Extras are already expanded to key / values by CKAN (unlike on
            package)
        * ~~Apply heuristic to unjsonify (if starts with [ or { unjsonify~~
        * JSON loads everything that starts with [ or {
    3. Map keys from CKAN to Frictionless (and reformat if needed)
    4. Remove keys with null values (CKAN has a lot of null valued keys)
    5. Apply special formatting (if any) for key fields e.g. slugiify
    '''
    # TODO: delete keys last as may be needed for something in processing
    resource = dict(ckandict)
    for key in resource_keys_to_remove:
        if key in resource:
            del resource[key]

    # unjsonify values
    # * check if string
    # * if starts with [ or { => json.loads it ...
    # HACK: bit of a hacky way to check if value is a jsonified array or
    # dict
    # * else do nothing
    for key, value in resource.items():
        if isinstance(value, str):
            value = value.strip()
            if value.startswith('{') or value.startswith('['):
                try:
                    value = json.loads(value)
                    resource[key] = value
                except (json_parse_exception, TypeError):
                    pass

    # Remap differences from CKAN to Frictionless resource
    for key, value in resource_mapping.items():
        if key in resource:
            resource[value] = resource[key]
            del resource[key]

    for key in list(resource.keys()):
        if resource[key] is None:
            del resource[key]

    return resource


dataset_keys_to_remove = [
    'state',          # b/c this is state info not metadata about dataset
    'isopen',         # computed info from license (render info not metadata)
    'num_resources',  # render info not metadata
    'num_tags',       # ditto
    'organization',   # already have owner_org id + this inlines related object
]
dataset_mapping = {
    'notes': 'description',
    'url': 'homepage'
}


def dataset(ckandict):
    '''Convert a CKAN Package (Dataset) to Frictionless Package.

    1. Expand extras.
        * JSON loads everything and on error have a string
    2. Map keys from CKAN to Frictionless (and reformat if needed)
    3. Remove keys with null values (CKAN has a lot of null valued keys)
    4. Remove unneeded keys
    5. Apply special formatting for key fields
    '''
    outdict = dict(ckandict)
    # Convert the structure of extras
    # structure of extra item is {key: xxx, value: xxx}
    if 'extras' in ckandict:
        for extra in ckandict['extras']:
            key = extra['key']
            value = extra['value']
            try:
                value = json.loads(value)
            except (json_parse_exception, TypeError):
                pass
            outdict[key] = value
        del outdict['extras']

    # Map dataset keys
    for key, value in dataset_mapping.items():
        if key in ckandict:
            outdict[value] = ckandict[key]
            del outdict[key]

    # map resources inside dataset
    if 'resources' in ckandict:
        outdict['resources'] = [resource(res) for res in
                                ckandict['resources']]

    # tags
    if ckandict.get('tags'):
        outdict['keywords'] = [tag['name'] for tag in ckandict['tags']]
    outdict.pop('tags', None)

    # author, maintainer => contributors
    # what to do if contributors already there? Options:
    # 1. Just use that and ignore author/maintainer
    # 2. replace with author/maintainer
    # 3. merge i.e. use contributors and merge in (this is sort of complex)
    # e.g. how to i avoid duplicating the same person
    # ANS: for now, is 1 ...
    if (not ('contributors' in outdict and outdict['contributors']) and
            ('author' in outdict or 'maintainer' in outdict)):
        outdict['contributors'] = []
        if 'author' in outdict and outdict['author']:
            contrib = {
                'title': outdict['author'],
                'role': 'author'
            }
            if 'author_email' in outdict:
                contrib['email'] = outdict['author_email']
            outdict['contributors'].append(contrib)
        if 'maintainer' in outdict and outdict['maintainer']:
            contrib = {
                'title': outdict['maintainer'],
                'role': 'maintainer'
            }
            if 'maintainer_email' in outdict:
                contrib['email'] = outdict['maintainer_email']
            outdict['contributors'].append(contrib)

    for key in ['author', 'author_email', 'maintainer', 'maintainer_email']:
        outdict.pop(key, None)

    # Algorithm for licenses
    # 1. Use extras first
    # 2. Updating first item in licenses array (if already there -
    # or create it as empty) with stuff at root of ckan dict i.e.
    # values from license_id, license_title etc.

    # Looping like this because all those keys are optional according to the
    # docs (though usually license_id will be there if others are there).
    for key in ['license_id', 'license_title', 'license_url']:
        if key in outdict and 'licenses' not in outdict:
            outdict['licenses'] = [{}]
            break  # check to create list of dicts only once
    if 'license_id' in outdict:
        outdict['licenses'][0]['name'] = outdict['license_id']
        outdict.pop('license_id', None)
    if 'license_title' in outdict:
        outdict['licenses'][0]['title'] = outdict['license_title']
        outdict.pop('license_title', None)
    if 'license_url' in outdict:
        outdict['licenses'][0]['path'] = outdict['license_url']
        outdict.pop('license_url', None)

    for key in dataset_keys_to_remove:
        outdict.pop(key, None)

    for key in list(outdict.keys()):
        if outdict[key] is None:
            del outdict[key]

    return outdict



import click
from ckanapi import RemoteCKAN
from dpckan.validations import run_dataset_validations
from dpckan.functions import (load_complete_datapackage, 
                              dataset_update, 
                              update_datapackage_json_resource,
                              dataset_path,
                              )
                              
def update(ckan_host, ckan_key, datapackage, datastore, exit_code):
  local_datapackage = load_complete_datapackage(datapackage)
  dataset = dataset_path(ckan_host, local_datapackage)
  click.echo(f"Updating dataset {dataset}")
  ckan_instance = RemoteCKAN(ckan_host, apikey = ckan_key)
  try:
    # run_resource_validations(ckan_instance, local_package)
    dataset_update(ckan_instance, local_datapackage, datastore, exit_code)
    click.echo(f"Dataset {dataset} updated")
  except Exception:
    click.echo(f"Error during {dataset} updating")
    if exit_code == True:
      sys.exit(1)

@click.command(name='update')
@click.pass_context
def update_cli(ctx):
  """
  Update dataset in a CKAN instance
  """
  update(
         ctx.obj['CKAN_HOST'], 
         ctx.obj['CKAN_KEY'], 
         ctx.obj['DATAPACKAGE'], 
         ctx.obj['DATASTORE'],
         ctx.obj['EXIT_CODE'],
         )



import sys
import os
import click
import hashlib
import json
from urllib.request import urlopen
from urllib import request
from frictionless_ckan_mapper import frictionless_to_ckan as f2c
from ckanapi import RemoteCKAN
from frictionless import Package
import shutil
import validators
import re

def load_complete_datapackage(source):
  datapackage = Package(source)
  for resource_name in datapackage.resource_names:
    datapackage.get_resource(resource_name).dialect.expand()
    datapackage.get_resource(resource_name).schema.expand()
  return datapackage

def dataset_create(ckan_instance, datapackage, datastore):
  # In this context datapackage is frictionless package object from datapackage.json local file
  remote_datapackage = frictionless_to_ckan(datapackage)
  ckan_instance.call_action('package_create', remote_datapackage)
  resources_ids = {}
  for resource_name in datapackage.resource_names:
    resource_ckan = resource_create(
                                    ckan_instance,
                                    datapackage.name,
                                    datapackage.get_resource(resource_name),
                                    datastore,
                                    )
    resources_ids[resource_name] = resource_ckan['id']
  ckan_datapackage_resource_id = create_datapackage_json_resource(ckan_instance, datapackage)
  resources_ids['datapackage.json'] = ckan_datapackage_resource_id
  datapackage['resources_ids'] = resources_ids
  dataset_patch_resources_ids(ckan_instance, datapackage)

def resource_create(ckan_instance, datapackage_id, resource, datastore):
  click.echo(f"Creating resource: {resource.name}")
  payload = {"package_id":datapackage_id,
               "name": resource.title,
               "description": resource.description,
               "url": resource.path}
  if(resource.path.startswith('http')):
    resource_ckan = ckan_instance.call_action('resource_create', payload)
  else:
    upload_files = {'upload': open(os.path.join(resource.basepath, resource.path), 'rb')}
    resource_ckan = ckan_instance.call_action('resource_create', payload, files=upload_files)
  if datastore == True:
    resource_update_datastore_metadata(ckan_instance,
                              resource_ckan['id'],
                              resource)
  return resource_ckan

def resource_update(ckan_instance, resource_id, resource, datastore):
  click.echo(f"Updating data and metadata of resource {resource.name}.")
  payload = {"id": resource_id,
             "name": resource.title,
             "description": resource.description,
             "url": resource.path}
  if(resource.path.startswith('http')):
    resource_ckan = ckan_instance.call_action('resource_update', payload)
  else:
    resource_ckan = ckan_instance.call_action('resource_update', payload,
                                       files={'upload': open(os.path.join(resource.basepath, resource.path), 'rb')})
  if datastore == True:
    resource_update_datastore_metadata(ckan_instance,
                                  resource_ckan['id'],
                                  resource)
  return resource_ckan

def resource_update_datastore_metadata(ckan_instance, resource_id, resource):
  click.echo(f"Updating resource {resource.name} datastore.")
  if resource.schema.fields == []:
    pass
  else:
    dataset_fields = {}
    resource_id = { "resource_id" : resource_id }
    dataset_fields.update(resource_id)
    force = { "force" : "True" }
    dataset_fields.update(force)
    fields = []
    for field in resource.schema.fields:
      meta_info = {"label": field.get("title", ""), "notes" : field.get("description", "") , "type_override" : 'text' }
      field = { "type" : 'text', "id" : field["name"] , "info" : meta_info }
      fields.append(field)
    dataset_fields.update({ "fields" : fields})
    ckan_instance.call_action('datastore_create', dataset_fields)

def delete_dataset(ckan_instance, dataset_name):
  ckan_instance.action.package_delete(id = dataset_name)

def is_dataset_published(ckan_instance, dataset_id):
  try:
    result = ckan_instance.action.package_show(id = dataset_id)
  except Exception:
    return False

  if(result['state'] == 'deleted'):
    return False

  return True

def create_datapackage_json_resource(ckan_instance, datapackage):
  click.echo("Criando recurso: datapackage.json")
  basepath = find_dataset_basepath(datapackage)
  expand_datapackage(datapackage, basepath)
  resource_ckan = ckan_instance.action.resource_create(package_id = datapackage.name,
                                       name = 'datapackage.json',
                                       upload = open(f"{basepath}/temp/datapackage.json", 'rb'))
  os.system(f'rm -rf {basepath}/temp')
  return resource_ckan['id']

def update_datapackage_json_resource(ckan_instance, datapackage, resource_id):
  click.echo(f"Updating resource datapackage.json.")
  datapackage.pop('resources_ids', None)
  basepath = find_dataset_basepath(datapackage)
  expand_datapackage(datapackage, basepath)
  ckan_instance.action.resource_update(id = resource_id,
                                       upload = open(f"{basepath}/temp/datapackage.json", 'rb'))
  os.system(f'rm -rf {basepath}/temp')

def find_dataset_basepath(datapackage):
  if datapackage.basepath == '':
    return '.'
  else:
    return datapackage.basepath

def expand_datapackage(datapackage, basepath):
  datapackage.to_json(f'{basepath}/temp/datapackage.json')

def dataset_update(ckan_instance, datapackage, datastore, exit_code):
  different_resources = dataset_diff(ckan_instance, datapackage)
  if len(different_resources) > 0:
    ckan_datapackage_resource_id = get_ckan_datapackage_resource_id(ckan_instance, datapackage.name)
    for resource in different_resources:
      if resource['name'] != 'datapackage.json':
        if resource['data_diff'] or resource['metadada_diff']:
          resource_update(ckan_instance, resource['id'], datapackage.get_resource(resource['name']), datastore)
          if datastore == True:
            resource_update_datastore_metadata(ckan_instance, resource['id'], datapackage.get_resource(resource['name']))
    update_datapackage_json_resource(ckan_instance, datapackage, ckan_datapackage_resource_id)
    dataset_patch(ckan_instance, datapackage)
  else:
    click.echo(f'Nothing to be updated in dataset {ckan_instance.address}/dataset/{datapackage.name}.')
    if exit_code == True:
      sys.exit(1)

def dataset_patch(ckan_instance, datapackage):
  datapackage['resources_ids'] = get_ckan_dataset_resources_ids(ckan_instance, datapackage)
  ckan_datapackage = frictionless_to_ckan(datapackage)
  ckan_datapackage['id'] = datapackage.name
  ckan_instance.call_action('package_patch', ckan_datapackage)

def dataset_diff(ckan_instance, datapackage):
  different_resources = list()
  ckan_dataset_resources_ids = get_ckan_dataset_resources_ids(ckan_instance, datapackage)
  ckan_dataset_resources_names = get_ckan_dataset_resources_names(ckan_dataset_resources_ids)
  for resource_name in ckan_dataset_resources_names:
    if resource_name in datapackage.resource_names or resource_name == 'datapackage.json':
      resource_id = ckan_dataset_resources_ids[resource_name]
      resource_diff = get_resource_diff(ckan_instance, datapackage, resource_name, resource_id)
      if resource_diff['data_diff'] or resource_diff['metadada_diff']:
        different_resources.append(resource_diff)
  return different_resources

def get_ckan_dataset_resources_ids(ckan_instance, datapackage):
  ckan_dataset = ckan_instance.action.package_show(id = datapackage.name)
  ckan_dataset_extras_property = ckan_dataset.get('extras')
  ckan_dataset_resources_ids = [i.get('value') for i in ckan_dataset_extras_property if i.get('key') == 'resources_ids']
  if len(ckan_dataset_resources_ids) == 0:
    click.echo(f"'resources_ids' property not found in 'extras' field of dataset {ckan_instance.address}/dataset/{datapackage.name}.")
    sys.exit(1)
  elif len(ckan_dataset_resources_ids) > 0:
    ckan_dataset_resources_ids = json.loads(ckan_dataset_resources_ids[0])
  return ckan_dataset_resources_ids

def get_ckan_dataset_resources_names(ckan_dataset_resources_ids):
  return [*ckan_dataset_resources_ids.keys()]

def get_resource_diff(ckan_instance, datapackage, resource_name, resource_id):
  resource_diff = dict()
  resource_diff['id'] = resource_id
  resource_diff['name'] = resource_name
  resource_diff['data_diff'] = is_resource_data_diff(ckan_instance,
                                                     datapackage,
                                                     resource_name,
                                                     resource_id)
  if resource_name != 'datapackage.json':
    resource_diff['metadada_diff'] = is_resource_metadata_diff(ckan_instance,
                                                       datapackage,
                                                       resource_name,
                                                       resource_id)
  else:
    resource_diff['metadada_diff'] = False
  return resource_diff

def is_resource_data_diff(ckan_instance, datapackage, resource_name, resource_id):
  local_data_hash = resource_hash(datapackage, resource_name)
  ckan_data_hash = resource_url_hash(ckan_instance, resource_id)
  if local_data_hash != ckan_data_hash:
    return True
  else:
    return False

def is_resource_metadata_diff(ckan_instance, datapackage, resource_name, resource_id):
  local_resource_metadata = datapackage.get_resource(resource_name)
  remote_dataset_metadata = get_remote_dataset_metadata(ckan_instance, datapackage.name)
  remote_resource_metadata = remote_dataset_metadata.get_resource(resource_name)
  if local_resource_metadata != remote_resource_metadata:
    return True
  else:
    return False

def get_remote_dataset_metadata(ckan_instance, dataset_name):
  ckan_datapackage_resource_id = get_ckan_datapackage_resource_id(ckan_instance, dataset_name)
  ckan_datapackage_resource = ckan_instance.action.resource_show(id=ckan_datapackage_resource_id)
  remote_dataset_metadata = Package(ckan_datapackage_resource['url'])
  return remote_dataset_metadata

def resource_hash(datapackage, name):
  resource_content = ''
  md5_hash = hashlib.md5()
  if name == 'datapackage.json':
    ckan_datapackage = frictionless_to_ckan(datapackage)
    resource_content = json.dumps(ckan_datapackage).encode('utf-8')
  else:
    basepath = find_dataset_basepath(datapackage)
    resource_path = datapackage.get_resource(name)['path']
    resource_content = open(f'{basepath}/{resource_path}', "rb").read()
  md5_hash.update(resource_content)
  resource_hash = md5_hash.hexdigest()
  return resource_hash

def resource_url_hash(ckan_instance, resource_id):
  resource_content = ''
  md5_hash = hashlib.md5()
  ckan_datapackage_resource = ckan_instance.action.resource_show(id = resource_id)
  if ckan_datapackage_resource['name'] == 'datapackage.json':
    # Buscar os metatados do dataset para retirar a key notes
    ckan_datapackage = ckan_instance.action.package_show(id=ckan_datapackage_resource['package_id'])
    # Buscar arquivo datapackage.json remoto para criar hash
    resource_content = urlopen(ckan_datapackage_resource['url']).read()
    resource_content = json.loads(resource_content.decode('utf-8'))
    resource_content = Package(resource_content)
    # Convert para metadados ckan para igualar à conversão do arquivo local
    # Esta conversão é importante para comparar modificações README, CHANGELOG e CONTRIBUTING
    resource_content = frictionless_to_ckan(resource_content)
    resource_content['notes'] = ckan_datapackage['notes']
    resource_content = json.dumps(resource_content).encode('utf-8')
  else:
    resource_content = urlopen(ckan_datapackage_resource['url']).read()
  md5_hash.update(resource_content)
  resource_hash = md5_hash.hexdigest()
  return resource_hash

def dataset_patch_resources_ids(ckan_instance, datapackage):
  ckan_datapackage = dict()
  ckan_datapackage['extras'] = frictionless_to_ckan(datapackage)['extras']
  ckan_datapackage['id'] = datapackage.name
  ckan_instance.call_action('package_patch', ckan_datapackage)

def frictionless_to_ckan(datapackage):
  dataset = f2c.package(datapackage)
  dataset.pop('resources') # Withdraw resources from dataset dictionary to avoid dataset creation with them
  README_path = os.path.join(datapackage.basepath, 'README.md')
  CONTRIBUTING_path = os.path.join(datapackage.basepath, 'CONTRIBUTING.md')
  CHANGELOG_path = os.path.join(datapackage.basepath, 'CHANGELOG.md')
  if "notes" not in dataset.keys():
    dataset["notes"] = ""
  if os.path.isfile(README_path):
    dataset["notes"] = ""
    dataset["notes"] += f"\n{open(README_path, encoding='utf-8').read()}"
  if os.path.isfile(CONTRIBUTING_path):
    dataset["notes"] += f"\n{open(CONTRIBUTING_path, encoding='utf-8').read()}"
  if os.path.isfile(CHANGELOG_path):
    dataset["notes"] += f"\n{open(CHANGELOG_path, encoding='utf-8').read()}"
  if 'id' in dataset.keys():
    dataset.update({ "id" : datapackage.name})
  return dataset

def get_ckan_datapackage_resource_id(ckan_instance, dataset_id):
  # Use show_package endpoint in ckan api to retrieve all dataset's resources
  ckan_datapackage_resources = ckan_instance.action.package_show(id=dataset_id)["resources"]
  # Filtering datackage_id - https://stackoverflow.com/a/48192370/11755155
  try:
    ckan_datapackage_resource_id = [i["id"] for i in ckan_datapackage_resources if i["url"].split('/')[-1] == "datapackage.json"][0]
    return ckan_datapackage_resource_id
  except IndexError:
    click.echo("Aborted: CKAN Dataset doesn't have required datapackage.json resource to complete the job.")
    sys.exit(1)

def get_dataset(ckan_host, link_id, path):
  if validate_url(link_id) and valid_dataset_url(link_id):
    ckan_host = split_dataset_url(link_id)[0]
    dataset_id = split_dataset_url(link_id)[1]
    download_dataset_resources(ckan_host, dataset_id, path)
  elif is_uuid(link_id):
    download_dataset_resources(ckan_host, link_id, path)
  elif not is_uuid(link_id) and valid_dataset_url(f"{ckan_host}/dataset/{link_id}"):
    download_dataset_resources(ckan_host, link_id, path)
  else:
    click.echo('Wrong arguments to get dataset. Chek dataset link or CKAN_HOST with dataset id or name informed.')
    sys.exit(1)

def download_dataset_resources(ckan_host, dataset_id, path):
  ckan_instance = RemoteCKAN(ckan_host)
  dataset_information = ckan_instance.action.package_show(id = dataset_id)
  dataset_name = dataset_information["name"]
  remote_dataset_metadata = get_remote_dataset_metadata(ckan_instance, dataset_name)
  path = path.split("/")
  path = list(filter(None, path))
  path = "/".join(path)
  if not os.path.exists(f'{path}/{dataset_name}'):
    click.echo(f'Creating {path}/{dataset_name} folder.')
    os.makedirs(f'{path}/{dataset_name}')
  else:
    if click.confirm(f'Folder {path}/{dataset_name} already exist. Do you want to delete it and continue?'):
      click.echo(f"Cleaning {path}/{dataset_name}'s folder contents.")
      shutil.rmtree(f'{path}/{dataset_name}')
      os.makedirs(f'{path}/{dataset_name}')
      for resource in dataset_information["resources"]:
        resource_url = resource["url"]
        if resource_url.split('/')[-1] == 'datapackage.json':
          click.echo('Downloading datapackage.json.')
          request.urlretrieve(resource_url, f'{path}/{dataset_name}/datapackage.json')
        else:
          file_name = resource_url.split('/')[-1].split('.')[0]
          file_path = remote_dataset_metadata.get_resource(file_name).path
          file_path = "/".join(file_path.split("/")[0:-1])
          os.makedirs(f'{path}/{dataset_name}/{file_path}', exist_ok=True)
          click.echo(f'Downloading {file_name} resource to {path}/{dataset_name}/{file_path}.')
          request.urlretrieve(resource_url, f'{path}/{dataset_name}/{file_path}/{file_name}.csv')
    else:
      click.echo('Process aborted.')

def split_dataset_url(dataset_url):
  dataset_url_list = list()
  dataset_url_splited = dataset_url.split('//')
  url_first_part = dataset_url_splited[0]
  url_second_part = dataset_url_splited[1].split('/')[0]
  host = f'{url_first_part}//{url_second_part}'
  dataset_id = dataset_url_splited[1].split('/')[2]
  dataset_url_list.append(host)
  dataset_url_list.append(dataset_id)
  return dataset_url_list

def valid_dataset_url(dataset_url):
  dataset_len = len(dataset_url.split('/'))
  if dataset_len >= 5:
    ckan_host = split_dataset_url(dataset_url)[0]
    dataset_id = split_dataset_url(dataset_url)[1]
    ckan_instance = RemoteCKAN(ckan_host)
    if is_dataset_published(ckan_instance, dataset_id):
      return True
    else:
      return False

def validate_url(url):
  response = validators.url(url)
  if response:
    return True
  else:
    return False

def is_uuid(uuid):
  matched = re.match('[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}', uuid)
  if bool(matched) == True:
    return True
  else:
    return False

def dataset_path(ckan_host, datapackage):
  dataset = f"{ckan_host}/dataset/{datapackage.name}"
  return dataset


import os.path

from ckanapi.errors import CKANAPIError
from ckanapi.common import (ActionShortcut, prepare_action,
    reverse_apicontroller_action)

class TestAppCKAN(object):
    """
    An interface to the the CKAN API actions on a paste TestApp

    :param test_app: the paste.fixture.TestApp instance, stored as
                    self.test_app
    :param apikey: the API key to pass as an 'X-CKAN-API-Key' header
                    when actions are called, stored as self.apikey
    """
    def __init__(self, test_app, apikey=None):
        self.test_app = test_app
        self.apikey = apikey
        self.action = ActionShortcut(self)

    def call_action(self, action, data_dict=None, context=None, apikey=None,
            files=None):
        """
        :param action: the action name, e.g. 'package_create'
        :param data_dict: the dict to pass to the action as JSON,
                          defaults to {}
        :param context: not supported
        :param files: None or {field-name: file-to-be-sent, ...}

        This function parses the response from the server as JSON and
        returns the decoded value.  When an error is returned this
        function will convert it back to an exception that matches the
        one the action function itself raised.
        """
        if context:
            raise CKANAPIError("TestAppCKAN.call_action does not support "
                "use of context parameter, use apikey instead")
        url, data, headers = prepare_action(action, data_dict,
                                            apikey or self.apikey, files)

        kwargs = {}
        if files:
            # Convert the list of (fieldname, file_object) tuples into the
            # (fieldname, filename, file_contents) tuples that webtests needs.
            upload_files = []
            for fieldname, file_ in files.items():
                if hasattr(file_, 'name'):
                    filename = os.path.split(file_.name)[1]
                else:
                    filename = fieldname
                upload_files.append( (fieldname, filename, file_.read()) )
            kwargs['upload_files'] = upload_files

        r = self.test_app.post('/' + url, data, headers, expect_errors=True,
                               **kwargs)
        return reverse_apicontroller_action(url, r.status, r.body)



import sys
import click
from ckanapi import RemoteCKAN
from dpckan.validations import run_resource_validations
from dpckan.functions import (
                              load_complete_datapackage,
                              update_datapackage_json_resource, 
                              resource_update,
                              get_ckan_datapackage_resource_id,
                              dataset_path,
                              )

def update_resource(ckan_host, ckan_key, datapackage, datastore, exit_code, resource_name, resource_id):
  local_package = load_complete_datapackage(datapackage)
  dataset = dataset_path(ckan_host, local_package)
  click.echo(f"Updating resource on {dataset}")
  ckan_instance = RemoteCKAN(ckan_host, apikey = ckan_key)
  try:
    run_resource_validations(ckan_instance, local_package)
    resource_update(ckan_instance,
                    resource_id,
                    local_package.get_resource(resource_name),
                    datastore,
                    )
    # Find ckan datapackage resource id
    ckan_datapackage_resource_id = get_ckan_datapackage_resource_id(ckan_instance, local_package.name)
    # Update ckan datapackage resource
    update_datapackage_json_resource(ckan_instance, local_package, ckan_datapackage_resource_id)
    click.echo(f"Resource {resource_name} updated on {dataset}")
  except Exception:
    print(f"Error updating {resource_name} resource on {dataset}")    
    if exit_code == True:
      sys.exit(1)

@click.command(name='update')
@click.argument('resource_id', required=True)
@click.pass_context
def update_resource_cli(ctx, resource_id):
  """
  Update dataset resource in a CKAN instance.
  """
  update_resource(
                  ctx.obj['CKAN_HOST'],
                  ctx.obj['CKAN_KEY'],
                  ctx.obj['DATAPACKAGE'],
                  ctx.obj['DATASTORE'],
                  ctx.obj['EXIT_CODE'],
                  ctx.obj['RESOURCE_NAME'],
                  resource_id,
                  )
  



from cgi import FieldStorage
from tempfile import TemporaryFile

from ckanapi.errors import CKANAPIError
from ckanapi.common import ActionShortcut

COPY_CHUNK = 1024*1024

class LocalCKAN(object):
    """
    An interface to calling actions with get_action() for CKAN plugins.

    :param username: perform action as this user, defaults to the site user
                     and stored as self.username
    :param context: a default context dict to use when calling actions,
                    stored as self.context with username added as its 'user'
                    value
    """
    def __init__(self, username=None, context=None):
        from ckan.logic import get_action
        self._get_action = get_action

        if username is None:
            username = self.get_site_username()
        self.username = username
        self.context = dict(context or [], user=self.username)
        self.action = ActionShortcut(self)

    def get_site_username(self):
        user = self._get_action('get_site_user')({'ignore_auth': True}, ())
        return user['name']

    def call_action(self, action, data_dict=None, context=None, apikey=None,
            files=None):
        """
        :param action: the action name, e.g. 'package_create'
        :param data_dict: the dict to pass to the action, defaults to {}
        :param context: an override for the context to use for this action,
                        remember to include a 'user' when necessary
        :param apikey: not supported
        :param files: None or {field-name: file-to-be-sent, ...}
        """
        # copy dicts because actions may modify the dicts they are passed
        # (CKAN...you so crazy)
        data_dict = dict(data_dict or [])
        context = dict(self.context if context is None else context)
        if apikey:
            # FIXME: allow use of apikey to set a user in context?
            raise CKANAPIError("LocalCKAN.call_action does not support "
                "use of apikey parameter, use context['user'] instead")

        to_close = []
        try:
            for fieldname in files or []:
                f = files[fieldname]
                if isinstance(f, tuple):
                    # requests accepts (filename, file...) tuples
                    filename, f = f[:2]
                else:
                    filename = f.name
                try:
                    f.seek(0)
                except (AttributeError, IOError):
                    f = _write_temp_file(f)
                    to_close.append(f)
                field_storage = FieldStorage()
                field_storage.file = f
                field_storage.filename = filename
                data_dict[fieldname] = field_storage

            return self._get_action(action)(context, data_dict)
        finally:
            for f in to_close:
                f.close()


def _write_temp_file(f):
    """
    Pull all data from stream f into a temporary file

    Caller must close file returned.
    """
    out = TemporaryFile()
    while True:  # FIXME: check for maximum size?
        chunk = f.read(COPY_CHUNK)
        if not chunk:
            break
        out.write(chunk)
    return out




import json
from io import BytesIO
try:
    import unittest2 as unittest
except ImportError:
    import unittest

import ckanapi


UPLOAD_DATA = b"""
public info
""".lstrip()

def wsgi_app(environ, start_response):
    status = '200 OK'
    headers = [('Content-type', 'application/json')]

    path = environ['PATH_INFO']
    if path == '/api/action/hello_world':
        response = {'success': True, 'result': 'how are you?'}
    elif path == '/api/action/invalid':
        response = {'success': False, 'error': {'__type': 'Validation Error'}}
    elif path == '/api/action/echo':
        content = environ['wsgi.input'].read()
        response = {'success': True, 'result':
            json.loads(content)['message']}
    elif path == '/api/action/upload':
        # poor man's multipart parsing
        data = environ['wsgi.input'].read().split('\r\n')
        saw_file = False
        for i, s in enumerate(data):
            if saw_file:
                if s == b'':
                    break
            elif b'filename="f"' in s:
                saw_file = True
        response = {'success': True, 'result': data[i + 1].decode('ascii')}
    elif path == '/api/action/not_ckan':
        response = {'success': False, 'error': True}

    start_response(status, headers)
    return [json.dumps(response)]


class TestTestAPPCKAN(unittest.TestCase):
    def setUp(self):
        try:
            import paste.fixture
        except ImportError:
            raise unittest.SkipTest('paste not importable')
        self.test_app = paste.fixture.TestApp(wsgi_app)
        self.ckan = ckanapi.TestAppCKAN(self.test_app)

    def test_simple(self):
        self.assertEqual(
            self.ckan.action.hello_world(), 'how are you?')

    def test_invalid(self):
        self.assertRaises(
            ckanapi.ValidationError,
            self.ckan.action.invalid)

    def test_data(self):
        self.assertEqual(
            self.ckan.action.echo(message='for you'), 'for you')

    def test_upload_action(self):
        self.assertEqual(
            self.ckan.action.upload(package_id='42',
                f=BytesIO(UPLOAD_DATA)), 'public info\n')

    def test_not_ckan(self):
        self.assertRaises(
            ckanapi.ServerIncompatibleError,
            self.ckan.action.not_ckan)



try:
    from urllib2 import Request, urlopen, HTTPError
    from urlparse import urlparse
except ImportError:
    from urllib.request import Request, urlopen, HTTPError
    from urllib.parse import urlparse

from ckanapi.errors import CKANAPIError
from ckanapi.common import (ActionShortcut, prepare_action,
    reverse_apicontroller_action)
from ckanapi.version import __version__

# add your sites here to remove parallel limits on ckanapi cli
MY_SITES = ['localhost', '127.0.0.1', '[::1]']

# add your site above instead of changing this
PARALLEL_LIMIT = 3

import requests


class RemoteCKAN(object):
    """
    An interface to the the CKAN API actions on a remote CKAN instance.

    :param address: the web address of the CKAN instance, e.g.
                    'http://demo.ckan.org', stored as self.address
    :param apikey: the API key to pass as an 'X-CKAN-API-Key' header
                    when actions are called, stored as self.apikey
    :param user_agent: the User-agent to report when making requests
    :param get_only: only use GET requests (default: False)
    """
    def __init__(self, address, apikey=None, user_agent=None, get_only=False):
        self.address = address
        self.apikey = apikey
        self.get_only = get_only
        self.session = None
        if not user_agent:
            user_agent = "ckanapi/{version} (+{url})".format(
                version=__version__,
                url='https://github.com/ckan/ckanapi')
        self.user_agent = user_agent
        self.action = ActionShortcut(self)

        net_loc = urlparse(address)
        if ']' in net_loc:
            net_loc = net_loc[:net_loc.index(']') + 1]
        elif ':' in net_loc:
            net_loc = net_loc[:net_loc.index(':')]
        if net_loc not in MY_SITES:
            # add your sites to MY_SITES above instead of removing this
            self.parallel_limit = PARALLEL_LIMIT

    def call_action(self, action, data_dict=None, context=None, apikey=None,
            files=None, requests_kwargs=None):
        """
        :param action: the action name, e.g. 'package_create'
        :param data_dict: the dict to pass to the action as JSON,
                          defaults to {}
        :param context: always set to None for RemoteCKAN
        :param apikey: API key for authentication
        :param files: None or {field-name: file-to-be-sent, ...}

        This function parses the response from the server as JSON and
        returns the decoded value.  When an error is returned this
        function will convert it back to an exception that matches the
        one the action function itself raised.
        """
        if context:
            raise CKANAPIError("RemoteCKAN.call_action does not support "
                "use of context parameter, use apikey instead")
        if files and self.get_only:
            raise CKANAPIError("RemoteCKAN: files may not be sent when "
                "get_only is True")
        url, data, headers = prepare_action(
            action, data_dict, apikey or self.apikey, files)
        headers['User-Agent'] = self.user_agent
        url = self.address.rstrip('/') + '/' + url
        requests_kwargs = requests_kwargs or {}
        if not self.session:
            self.session = requests.Session()
        if self.get_only:
            status, response = self._request_fn_get(url, data_dict, headers, requests_kwargs)
        else:
            status, response = self._request_fn(url, data, headers, files, requests_kwargs)
        return reverse_apicontroller_action(url, status, response)

    def _request_fn(self, url, data, headers, files, requests_kwargs):
        r = self.session.post(url, data=data, headers=headers, files=files,
            allow_redirects=False, **requests_kwargs)
        # allow_redirects=False because: if a post is redirected (e.g. 301 due
        # to a http to https redirect), then the second request is made to the
        # new URL, but *without* the data. This gives a confusing "No request
        # body data" error. It is better to just return the 301 to the user, so
        # we disallow redirects.
        return r.status_code, r.text

    def _request_fn_get(self, url, data_dict, headers, requests_kwargs):
        r = self.session.get(url, params=data_dict, headers=headers,
            **requests_kwargs)
        return r.status_code, r.text

    def close(self):
        """Close session"""
        if self.session:
            self.session.close()
            self.session = None

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()



import json
import cgi
import csv
from wsgiref.util import setup_testing_defaults
from wsgiref.simple_server import make_server

def mock_ckan(environ, start_response):
    status = '200 OK'
    headers = [
        ('Content-type', 'application/json;charset=utf-8'),
        ]
    if environ['PATH_INFO'] == '/api/action/site_read':
        start_response(status, headers)
        return [json.dumps(True).encode('utf-8')]
    if environ['PATH_INFO'] == '/api/action/organization_list':
        start_response(status, headers)
        return [json.dumps({
            "help": "none",
            "success": True,
            "result": ["aa", "bb", "cc"]
            }).encode('utf-8')]
    if environ['PATH_INFO'] == '/api/action/test_echo_user_agent':
        start_response(status, headers)
        return [json.dumps({
            "help": "none",
            "success": True,
            "result": environ['HTTP_USER_AGENT']
            }).encode('utf-8')]
    if environ['PATH_INFO'] == '/api/action/test_echo_content_type':
        start_response(status, headers)
        return [json.dumps({
            "help": "none",
            "success": True,
            "result": environ['CONTENT_TYPE']
            }).encode('utf-8')]
    if environ['PATH_INFO'] == '/api/action/test_upload':
        fs = cgi.FieldStorage(
            fp=environ['wsgi.input'],
            environ=environ,
            keep_blank_values=True,
            )
        records = list(csv.reader(fs['upload'].file))
        start_response(status, headers)
        return [json.dumps({
            "help": "none",
            "success": True,
            "result": {
                'option': fs.getvalue('option'),
                'last_row': records[-1],
                },
            }).encode('utf-8')]
    if environ['PATH_INFO'].startswith('/api/action/'):
        start_response(status, headers)
        return [json.dumps({
            "help": "none",
            "success": False,
            "error": {'__type': 'Not Found Error'},
            }).encode('utf-8')]
    start_response('404 Not Found', headers)
    return []

httpd = make_server('localhost', 8901, mock_ckan)
httpd.serve_forever()




import os
import click
from ckanapi import RemoteCKAN
from urllib import request
from dpckan.functions import get_dataset

@click.command(name='get')
@click.argument('url', required=True)
@click.argument('path', default='.')
@click.pass_context
def get_dataset_cli(ctx, url, path):
  """
  Get a dataset published in ckan.

  URL required argument could be a Link or name or id of the wanted dataset
  """
  get_dataset(ctx.obj['CKAN_HOST'], url, path)



from ckanapi.cli.workers import worker_pool
import os

try:
    import unittest2 as unittest
except ImportError:
    import unittest


class _MockPopen(object):
    def __init__(self, popen_args, stdin, stdout):
        read1fd, write1fd = os.pipe()
        read2fd, write2fd = os.pipe()
        self.stdin = os.fdopen(write1fd, 'wb')
        self.stdin_inside = os.fdopen(read1fd, 'rb')
        self.stdout = os.fdopen(read2fd, 'rb')
        self.stdout_inside = os.fdopen(write2fd, 'wb')
        # use popen_args as an after-create callback
        popen_args(self)

    def stdout_write(self, data):
        self.stdout_inside.write(data)
        self.stdout_inside.flush()

    def stdin_readline(self):
        return self.stdin_inside.readline()

    def close_pipes(self):
        for f in (self.stdin, self.stdin_inside, self.stdout, self.stdout_inside):
            f.close()


class TestCLIWorkers(unittest.TestCase):
    def test_one(self):
        children = []
        def child_created(child):
            # need to respond or pool will block test
            child.stdout_write(b'AA\n')
            children.append(child)
        pool = worker_pool(
            child_created,
            1,
            enumerate((b"job1\n", b"job2\n")),
            popen=_MockPopen,
            )
        response = next(pool)
        self.assertEqual(len(children), 1)
        c = children[0]
        self.assertEqual(c.stdin_readline(), b'job1\n')
        self.assertEqual(response, ([1], 0, b'AA\n'))
        self.assertEqual(c.stdin_readline(), b'job2\n')
        c.stdout_write(b'BB\n')
        self.assertEqual(next(pool), ([None], 1, b'BB\n'))
        self.assertRaises(StopIteration, next, pool)
        for c in children:
            c.close_pipes()

    def test_two(self):
        children = []
        def child_created(child):
            # first child responds
            if not children:
                child.stdout_write(b'AA\n')
            children.append(child)
        pool = worker_pool(
            child_created,
            2,
            enumerate((b"job1\n", b"job2\n", b"job3\n", b"job4\n")),
            popen=_MockPopen,
            )
        response = next(pool)
        self.assertEqual(len(children), 2)
        c0, c1 = children
        self.assertEqual(c0.stdin_readline(), b'job1\n')
        self.assertEqual(c1.stdin_readline(), b'job2\n')
        self.assertEqual(response, ([2, 1], 0, b'AA\n'))
        self.assertEqual(c0.stdin_readline(), b'job3\n')
        c1.stdout_write(b'BB\n')
        self.assertEqual(next(pool), ([2, 3], 1, b'BB\n'))
        self.assertEqual(c1.stdin_readline(), b'job4\n')
        c0.stdout_write(b'CC\n')
        self.assertEqual(next(pool), ([None, 3], 2, b'CC\n'))
        c1.stdout_write(b'DD\n')
        self.assertEqual(next(pool), ([None, None], 3, b'DD\n'))
        self.assertRaises(StopIteration, next, pool)
        for c in children:
            c.close_pipes()

    def test_uneven(self):
        children = []
        def child_created(child):
            # second child responds
            if children:
                child.stdout_write(b'AA\n')
            children.append(child)
        pool = worker_pool(
            child_created,
            2,
            enumerate((b"job1\n", b"job2\n", b"job3\n", b"job4\n")),
            popen=_MockPopen,
            )
        response = next(pool)
        self.assertEqual(len(children), 2)
        c0, c1 = children
        self.assertEqual(c0.stdin_readline(), b'job1\n')
        self.assertEqual(c1.stdin_readline(), b'job2\n')
        self.assertEqual(response, ([0, 2], 1, b'AA\n'))
        self.assertEqual(c1.stdin_readline(), b'job3\n')
        c1.stdout_write(b'BB\n')
        self.assertEqual(next(pool), ([0, 3], 2, b'BB\n'))
        self.assertEqual(c1.stdin_readline(), b'job4\n')
        c1.stdout_write(b'CC\n')
        self.assertEqual(next(pool), ([0, None], 3, b'CC\n'))
        c0.stdout_write(b'DD\n')
        self.assertEqual(next(pool), ([None, None], 0, b'DD\n'))
        self.assertRaises(StopIteration, next, pool)
        for c in children:
            c.close_pipes()

    def test_overkill(self):
        children = []
        def child_created(child):
            if not children:
                child.stdout_write(b'AA\n')
            children.append(child)
        pool = worker_pool(
            child_created,
            10,
            enumerate((b"job1\n",)),
            popen=_MockPopen,
            )
        response = next(pool)
        self.assertEqual(len(children), 1)
        c = children[0]
        self.assertEqual(c.stdin_readline(), b'job1\n')
        self.assertEqual(response, ([None], 0, b'AA\n'))
        self.assertRaises(StopIteration, next, pool)
        for c in children:
            c.close_pipes()

    def test_batch(self):
        children = []
        def child_created(child):
            if not children:
                child.stdout_write(b'AA\n')
            children.append(child)
        pool = worker_pool(
            child_created,
            2,
            enumerate((b"job1\n",)),
            stop_when_jobs_done=False,
            popen=_MockPopen,
            )
        response = next(pool)
        self.assertEqual(len(children), 1)
        c0 = children[0]
        self.assertEqual(c0.stdin_readline(), b'job1\n')
        self.assertEqual(response, ([None], 0, b'AA\n'))
        self.assertEqual(next(pool), (None, None, None))
        # need to write in advance to avoid blocking test
        c0.stdout_write(b'BB\n')
        response = pool.send(enumerate((b"job2\n", b"job3\n"), 1))
        self.assertEqual(response, ([None, 2], 1, b'BB\n'))
        for c in children:
            c.close_pipes()



from pathlib import Path
from dotenv import load_dotenv
from dpckan import create_dataset
from dpckan import update_dataset
from dpckan import create_resource
from dpckan import update_resource

load_dotenv(dotenv_path=Path('.', '.env'))


